from typing import Callable

def _create_function(names:list[str], ls_couplings:dict[int, dict[str: dict[tuple, float]]], f, complex_couplings=False) -> tuple[Callable, list[str]]:
    from decayamplitude.resonance import LSTuple, Resonance
    import inspect
    import types
    # Create a function signature dynamically
    
    coupling_names = []
    coupling_structure = {}
    for resonance_id, coupling_dict in ls_couplings.items():
        coupling_structure[resonance_id] = {}
        for key, _ in coupling_dict["couplings"].items():
            resonance = Resonance.get_instance(resonance_id)
            name = f"{resonance.descriptor}_{'LS' if resonance.scheme == 'ls' else 'H'}_{'_'.join([str(k) for k in key])}"
            if complex_couplings:
                name_real = f"{name}_real"
                name_imag = f"{name}_imaginary"
                coupling_names.append(name_real)
                coupling_names.append(name_imag)
            else:
                coupling_names.append(name) # we need only define a name 
            coupling_structure[resonance_id][key] = name
    full_names = names + coupling_names
    names_with_duplicates = full_names.copy()
    full_names = list(set(full_names)) # remove duplicates, since the same decay process can exist in multiple chains
    # Sort the names to ensure consistent ordering as given from the outside
    full_names.sort(key=lambda x: names_with_duplicates.index(x))
    # Define a generic function that accepts *args
    def func(*args, **kwargs):
        named_map = {name: arg for name, arg in zip(full_names, args)}
        named_map.update(kwargs)
        couplings = {}

        def compute_coupling(name):
            if complex_couplings:
                return named_map[f"{name}_real"] + 1j * named_map[f"{name}_imaginary"]
            else:
                return named_map[name]

        for resonance_id, coupling_dict in coupling_structure.items():

            couplings[resonance_id] = {"couplings":{
                key: compute_coupling(coupling_dict[key]) for key in coupling_dict
            }}
        arguments = named_map.copy()
        arguments.update(couplings)
        return f(arguments)

    # Assign the generated signature to the function
    # we use the set to remove duplicates. These can exist, if the same decay process exists in multiple chains
    parameters = [inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD) for name in full_names]
    sig = inspect.Signature(parameters)
    func.__signature__ = sig
    return func, full_names

def sanitize(name: str) -> str:
    """
    Sanitize a name for use in python code
    """
    replacements = [
        ("*", "star"),
        ("(", ""),
        (")", ""),
        ("[", ""),
        ("]", ""),
        ("{", ""),
        ("}", ""),
    ] + [
        (a, "_") for a in " .,/\\^'\"~!?=>|&$#@%;:`´§°"
    ]
    for replacement in replacements:
        name = name.replace(*replacement)
    name = name.replace("+", "plus")
    name = name.replace("-", "minus")
    
    return name
    
    



# This file was generated by decaytreeedit-backend
# Copyright © 2025 Kai Habermann

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Author(s) shall be credited in any publication or presentation that uses this software, including but not limited to conference presentations, journal articles, and theses. The Authors shall be credited by reference to the decayangle and decayamplitude repositories, as well as the citation of the original paper: https://doi.org/10.1103/PhysRevD.111.056015

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 

from decayamplitude.resonance import Resonance, ResonanceDict
from decayamplitude.rotation import QN
from decayamplitude.chain import MultiChain
from decayamplitude.combiner import ChainCombiner
from decayamplitude.backend import numpy as np
from decayangle.decay_topology import Topology, Node
from decayamplitude.particle import Particle, DecaySetup

from decayangle.config import config as decayangle_config
from decayamplitude.kinematics_helpers import mass_from_node
decayangle_config.backend = "jax"
decayangle_config.sorting = "value"  # this makes sure, that the order of entries in a node is handeled internaly and we do not have to worry about it

from jax import jit, grad

def constant_lineshape(L, S, *args):
    #This is a dummy lineshape function, it does nothing, but is needed for the code to work
    # It is suited to model a mother resonance for example
    return 1.0

def BW_lineshape(mass):
    # usually the function is called for a fixed set of points in the phasespace. 
    # Thus we can give the mass values as fixed
    # This allows the compiler to optimize the function better
    # decayamplitude will provide l and s values per default for the resonances, but not the mass values
    def BW(l,s, m0, gamma):
        return 1 / (mass**2 - m0**2 + 1j * mass * gamma) 
    return BW

        
final_state_qn = {
    1: Particle(quantum_numbers=QN(2, -1), name="J/psi(1S)"), # name for orientation: J/psi(1S)
    2: Particle(quantum_numbers=QN(0, -1), name="pi+"), # name for orientation: pi+
    3: Particle(quantum_numbers=QN(0, -1), name="pi-"), # name for orientation: pi-
    4: Particle(quantum_numbers=QN(0, -1), name="pi+"), # name for orientation: pi+
    5: Particle(quantum_numbers=QN(0, -1), name="pi-"), # name for orientation: pi-
}

setup = DecaySetup(final_state_particles=final_state_qn)

topologies = setup.topologies  # This ensures correct symmetrization

def amplitude(momenta):
    # momenta has the form of 
    # {
    #  1: np.array(shape=(...,4)), with indices 0,1,2,3 for p_x,p_y,p_z,E
    #  2: np.array(shape=(...,4)), with indices 0,1,2,3 for p_x,p_y,p_z,E
    #  ...
    #}

    resonances = ResonanceDict({
        0:    [
            Resonance(Node(0), quantum_numbers=QN(0, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name='B0') 
            # Mother particle, set the QN to what they need to be for your decay. Usually a weak initial decay is assumed, so parity is not preserved. If this is not the case set it to True here or in the frontend.
                    # Replace the lineshape with a function of your liking, the call signatrue has to be f(L, S, *args), where the args will be exposed under the names you give in argnames
                    # The mass points at which to evaluate the lineshape are not given at runtime. Look at the bw_lineshpe for details
            ],
        
        (2, 3, 4, 5): [
        Resonance(Node((2, 3, 4, 5)), quantum_numbers=QN(6, -1), lineshape=BW_lineshape(mass_from_node(Node((2, 3, 4, 5)), momenta)), argnames=['rho_316900_mass', 'rho_316900_width'], preserve_partity=True, name='rho_3(1690)0')
        ],

        (4, 5): [
        Resonance(Node((4, 5)), quantum_numbers=QN(2, -1), lineshape=BW_lineshape(mass_from_node(Node((4, 5)), momenta)), argnames=['rho7700_mass', 'rho7700_width'], preserve_partity=True, name='rho(770)0')
        ],

        (2, 3): [
        Resonance(Node((2, 3)), quantum_numbers=QN(2, -1), lineshape=BW_lineshape(mass_from_node(Node((2, 3)), momenta)), argnames=['rho7700_mass', 'rho7700_width'], preserve_partity=True, name='rho(770)0')
        ],

        (1, 2, 3): [
        Resonance(Node((1, 2, 3)), quantum_numbers=QN(2, -1), lineshape=BW_lineshape(mass_from_node(Node((1, 2, 3)), momenta)), argnames=['psi2S_mass', 'psi2S_width'], preserve_partity=True, name='psi(2S)')
        ]
    })

    topologies = setup.filled_topologies(
        resonances=resonances
    )
    print(f"TOPOLOGIES: {topologies}")
    chains = [
        MultiChain(
            topology=topology,
            resonances = resonances,
            momenta = momenta,
            final_state_qn = final_state_qn
        ) for topology in topologies
    ]
    full = ChainCombiner(chains)

    # The unpolarized amplitude is the simplest one, and the default case in LHCb
    unpolarized, argnames = full.unpolarized_amplitude(
        full.generate_couplings() # This is a helper function to generate the couplings for the hadronic system, if you want to restrict them, you will have to do it manually.
                                    # Alternatively you can also restrict the couplings in the fitter later.      
        )
    # argnames are the names of the arguments of the function, which are the masses and widths of the resonances and the couplings
    # The order of argnames is the order of the arguments in the function
    # The function can be called with positional arguments, or with keyword arguments
    # so unpolarized(*[1, 2, 3, 4, 5, 6]) is the same as unpolarized(mass_resonance_1=1, width_resonance_1=2, mass_resonance_2=3, width_resonance_2=4, mass_resonance_3=5, width_resonance_3=6), omitting the couplings
    # print(argnames)
    for name in argnames:
        print(name)

    # an issue with jax, where the internal caching structure needs to be prewarmed, so that in the compilation step the correct types are inferred
    print(unpolarized(*([1.0] * len(argnames))))

    # we can now jit the function, to make it faster after the compile
    unpolarized = jit(unpolarized) 
    print(unpolarized(*([1.0] * len(argnames))))

    # for the gradient calculation we need to define a log likelihood function or something, that produces a single value
    def LL(*args):
        return np.sum(
            np.log(unpolarized(*args))
                )
    # we can calc the gradient of the log likelihood function
    unpolarized_grad = jit(grad(LL, argnums=[i for i in range(len(argnames))]))

    # and a test call (may take quite some time)
    # print(unpolarized_grad(*([1.0] * len(argnames))))

    # Further calls will be much faster, since we dont need to compile again
    # print(unpolarized_grad(*([1.0] * len(argnames))))
    # print(unpolarized_grad(*([2.0] * len(argnames))))


    # Other options for amplitudes, one might be interested in
    # polarized, lambdas ,polarized_argnames = full.polarized_amplitude(full.generate_couplings())
    # print(lambdas)
    # print(polarized(*lambda_values,*([1] * len(polarized_argnames))) )

    # matrix_function, matrix_argnames = full.matrix_function(full.generate_couplings())
    # print(matrix_argnames)
    # print(matrix_function(0, *([1] * len(argnames))) )

    return full

if __name__ == "__main__":
    mother_mass = 1
    daughter_masses = [0.1] * len(final_state_qn)
    from decayangle.kinematics import from_mass_and_momentum
    import random 

    random_angles = [
        (random.uniform(0, 2 * np.pi) , random.uniform(0, np.pi)) 
         for _ in range(len(final_state_qn))
         ]

    three_momenta = [
        np.array([0.1,0,0]) * np.array([np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)])
        for theta, phi in random_angles
    ]

    momenta = {
        i: from_mass_and_momentum(m, p)
            for i, (m, p) in enumerate(zip(daughter_masses, three_momenta), start=1)
    }
    momenta = topologies[0].to_rest_frame(momenta)
    full = amplitude(momenta)


from __future__ import annotations
from decayamplitude.rotation import QN
from decayamplitude.chain import MultiChain
from decayamplitude.combiner import ChainCombiner
from decayamplitude.resonance import Resonance
from decayangle.decay_topology import Topology, Node
from decayamplitude.kinematics_helpers import mass_from_node

import numpy as np
def constant_lineshape(*args):
    return 1.0


def test_multi_chain():
    momenta = {
        1: np.array([1, 0.1, 0.4, 3]),
        2: np.array([0.5, -0.1, -0.4, 3]),
        3: np.array([1.1, 0.2, 0.5, 3]),
        4: np.array([0.6, -0.2, -0.5, 3]),
    }
    final_state_qn = {
            1: QN(0, 1), 
            2: QN(0, 1), 
            3: QN(1, 1), 
            4: QN(1, -1) 
        }


    m = mass_from_node(Node((1,2,3)), momenta)
    resonances_hadronic = {
        (1,2): [
            # Here the hadronic resonances go+
            # These will decay strong, so we need to conserve parity
            Resonance(Node((1, 2)), quantum_numbers=QN(0, 1), lineshape=constant_lineshape, argnames=["D_2300_M", "D_2300_Gamma"], preserve_partity=True, name="Resnance1"),
            Resonance(Node((1, 2)), quantum_numbers=QN(4, 1), lineshape=constant_lineshape, argnames=["D_2460_M", "D_2460_Gamma"], preserve_partity=True, name="Resonance2"),
        ],
        (3, 4): 
        [Resonance(Node((3, 4)), quantum_numbers=QN(2, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="Resonance3")],
        (1,2,3): 
        [Resonance(Node((1, 2, 3)), quantum_numbers=QN(1, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="Resonance4")],

        0: [Resonance(Node(0), quantum_numbers=QN(0, 1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="B0")],
    }

    topology1 = Topology(
        0,
        decay_topology=((1,2), (3, 4))
    )

    momenta = topology1.to_rest_frame(momenta)

    topology2 = Topology(
        0,
        decay_topology=(((1,2), 3) ,4 )
    )

    chain1 = MultiChain(
        topology = topology1,
        resonances = resonances_hadronic,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    chain2 = MultiChain(
        topology = topology2,
        resonances = resonances_hadronic,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    assert len(chain1.chains) == 2
    for chain in chain1.chains:
        # root resonance + 2 chain resonances
        assert len(chain.resonances) == 3

    assert len(chain2.chains) == 2
    for chain in chain2.chains:
        assert len(chain.resonances) == 3

    combined = ChainCombiner([chain1, chain2])
    func, params = combined.unpolarized_amplitude(combined.generate_couplings())

if __name__ == "__main__":
    test_multi_chain()


from utils import make_four_vectors, constant_lineshape, BW_lineshape
from decayamplitude.resonance import Resonance
from decayamplitude.rotation import QN, Angular
from decayamplitude.chain import DecayChain, MultiChain
from decayamplitude.combiner import ChainCombiner

from decayamplitude.backend import numpy as np

from decayangle.decay_topology import Topology, Node, TopologyCollection
from decayangle.kinematics import mass
from decayangle.config import config as decayangle_config
decayangle_config.backend = "jax"
decayangle_config.sorting = "off"

from jax import jit, grad

from collections import defaultdict

# Since we are semi leptonic there is only one important decay topology
# The (1, 2) system is the hadronic system, which is the one we are interested in
# Decay definition: B0 -> pip pim (J/psi -> mu+ mu-) 

# this will enforce, that we only get topologies, which have the intermediate state J/Psi -> mu+ mu-
tc = TopologyCollection(0, [1,2,3,4])

topology_pipi, = tc.filter((3,4), (1,2))
topology_pip_JPsi, = tc.filter((3,4), (1, 3, 4))
topology_pi_m_JPsi, = tc.filter((3,4), (2, 3, 4))
def resonances(momenta):
    m_pi_pi = topology_pipi.nodes[(1,2)].mass(momenta=momenta)
    resonances_pi_pi = {
        (1,2): [
            # Here the pi pi resonances go
            # These will decay strong, so we need to conserve parity
            Resonance(Node((1, 2)), quantum_numbers=QN(2, 1), lineshape=BW_lineshape(m_pi_pi), argnames=["mass_resonance_1", "width_resonance_1"], preserve_partity=True, name="pipi_resonance_1"),
            ],
        # This is the J/Psi meson. It is defined as a resonance, but we assue a constant lineshape, since it is a extremely narrow resonance. 
        # By setting this resonance to the helicity scheme, the resulting function will expect helicity couplings and not ls couplings
        # Note, that the preserve_partity flag is still set to True, but does not have any effect in the helicity scheme, as here parity conservation can not be enforced by truncating the couplings
        # For the J/Psi decay, we know, that the helicity couplings are all the same and we can thus just fix them in the fitter.
        (3, 4): [Resonance(Node((3, 4)), quantum_numbers=QN(2, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=True, name="J/Psi_1", scheme="helicity")],

        # This is the decaying B0 meson. It is defined as a resonance, but since this is a decay amplitude, the description is not important. Only the QN have to be correct aswell as the parity conservation, since this controlls the automated ls couplings
        0: [Resonance(Node(0), quantum_numbers=QN(0, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="B0_1")]
    }

    m_pip_JPsi = topology_pip_JPsi.nodes[(1,3,4)].mass(momenta=momenta)
    resonances_pip_JPsi = {
        (1, 3, 4): [
            # Here the pip JPsi resonances go
            # These will decay strong, so we need to conserve parity
            Resonance(Node((1, 3, 4)), quantum_numbers=QN(2, -1), lineshape=BW_lineshape(m_pip_JPsi), argnames=["mass_resonance_2", "width_resonance_2"], preserve_partity=True, name="J/Psi_resonance_1"),
            ],
        # This is the J/Psi meson. It is defined as a resonance, but we assue a constant lineshape, since it is a extremely narrow resonance. 
        # By setting this resonance to the helicity scheme, the resulting function will expect helicity couplings and not ls couplings
        # Note, that the preserve_partity flag is still set to True, but does not have any effect in the helicity scheme, as here parity conservation can not be enforced by truncating the couplings
        # For the J/Psi decay, we know, that the helicity couplings are all the same and we can thus just fix them in the fitter.
        (3, 4): [Resonance(Node((3, 4)), quantum_numbers=QN(2, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=True, name="J/Psi_2", scheme="helicity")],

        # This is the decaying B0 meson. It is defined as a resonance, but since this is a decay amplitude, the description is not important. Only the QN have to be correct aswell as the parity conservation, since this controlls the automated ls couplings
        0: [Resonance(Node(0), quantum_numbers=QN(0, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="B0_2")]
    }

    m_pi_m_JPsi = topology_pi_m_JPsi.nodes[(2,3,4)].mass(momenta=momenta)
    resonances_pi_m_JPsi = {
        (2, 3, 4): [
            # Here the pi- JPsi resonances go
            # These will decay strong, so we need to conserve parity
            Resonance(Node((2, 3, 4)), quantum_numbers=QN(0, -1), lineshape=BW_lineshape(m_pi_m_JPsi), argnames=["mass_resonance_3", "width_resonance_3"], preserve_partity=True, name="J/Psi_resonance_2"),
            ],
        # This is the J/Psi meson. It is defined as a resonance, but we assue a constant lineshape, since it is a extremely narrow resonance. 
        # By setting this resonance to the helicity scheme, the resulting function will expect helicity couplings and not ls couplings
        # Note, that the preserve_partity flag is still set to True, but does not have any effect in the helicity scheme, as here parity conservation can not be enforced by truncating the couplings
        # For the J/Psi decay, we know, that the helicity couplings are all the same and we can thus just fix them in the fitter.
        (3, 4): [Resonance(Node((3, 4)), quantum_numbers=QN(2, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=True, name="J/Psi_3", scheme="helicity")],

        # This is the decaying B0 meson. It is defined as a resonance, but since this is a decay amplitude, the description is not important. Only the QN have to be correct aswell as the parity conservation, since this controlls the automated ls couplings
        0: [Resonance(Node(0), quantum_numbers=QN(0, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="B0_3")]
    }

    return resonances_pi_pi, resonances_pip_JPsi, resonances_pi_m_JPsi


def phasespace_momenta():
    import phasespace
    B0_MASS = 5.27963
    PION_MASS = 0.13957018
    MU_MASS = 0.1056583715

    weights, particles = phasespace.nbody_decay(B0_MASS,
                                            [PION_MASS, PION_MASS, MU_MASS, MU_MASS]).generate(n_events=10)
    momenta = {
        1: np.array(particles["p_0"]),
        2: np.array(particles["p_1"]),
        3: np.array(particles["p_2"]),
        4: np.array(particles["p_3"]),
    }
    return momenta

def shortFourBodyAmplitudeBW():
    final_state_qn = {
        1: QN(0, 1), # pi+
        2: QN(0, -1), # pi-
        3: QN(1, 1), # mu+
        4: QN(1, -1) # mu-
    }
    momenta = phasespace_momenta()

    resonances_pi_pi, resonances_pip_JPsi, resonances_pi_m_JPsi = resonances(momenta)
    chain1 = MultiChain(
        topology = topology_pipi,
        resonances = resonances_pi_pi,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    chain2 = MultiChain(
        topology = topology_pip_JPsi,
        resonances = resonances_pip_JPsi,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    chain3 = MultiChain(
        topology = topology_pi_m_JPsi,
        resonances = resonances_pi_m_JPsi,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    # For semileptonics no actual alignment is needed, as we are only interested in one single decay chain
    # But doing it this way allows to use the convenience functions of the combiner
    full = ChainCombiner([chain1, chain2, chain3])

    # The unpolarized amplitude is the simplest one, and the default case in LHCb
    unpolarized, argnames = full.unpolarized_amplitude(
        full.generate_couplings() # This is a helper function to generate the couplings for the hadronic system, if you want to restrict them, you will have to do it manually.
                                    # Alternatively you can also restrict the couplings in the fitter later.      
        )
    # argnames are the names of the arguments of the function, which are the masses and widths of the resonances and the couplings
    # The order of argnames is the order of the arguments in the function
    # The function can be called with positional arguments, or with keyword arguments
    # so unpolarized(*[1, 2, 3, 4, 5, 6]) is the same as unpolarized(mass_resonance_1=1, width_resonance_1=2, mass_resonance_2=3, width_resonance_2=4, mass_resonance_3=5, width_resonance_3=6), omitting the couplings
    print(argnames)

    # an issue with jax, where the internal caching structure needs to be prewarmed, so that in the compilation step everything, that can be static, is static
    print(unpolarized(*([1] * len(argnames))))
    
    # we can now jit the function
    unpolarized = jit(unpolarized) 
    print(unpolarized(*([1] * len(argnames))))


    # for the gradient calculation we need to define a log likelihood function or something, that produces a single value
    def LL(*args):
        # Warning: This is not the correct log likelihood function, but just a dummy
        # The amplitudes do not come normalized, so this can not be used for fitting directly
        return np.sum(
            np.log(unpolarized(*args))
                )
    # we can calc the gradient of the log likelihood function
    unpolarized_grad = jit(grad(LL, argnums=[i for i in range(len(argnames))]))

    # and a test call
    print(unpolarized_grad(*([1.0] * len(argnames))))


    # Other options for amplitudes, one might be interested int
    # polarized, lambdas ,polarized_argnames = full.polarized_amplitude(full.generate_couplings())
    # print(lambdas)
    # lambda_values = [0, 0, 0, 1, 1]
    # print(polarized(*lambda_values,*([1] * len(polarized_argnames))) )

    # matrix_function, matrix_argnames = full.matrix_function(full.generate_couplings())
    # print(matrix_argnames)
    # print(matrix_function(0, *([1] * len(argnames))) )


if __name__ == "__main__":
    shortFourBodyAmplitudeBW()


from decayamplitude.backend import numpy as np
import decayamplitude
from decayamplitude.chain import DecayChain, MultiChain, AlignedChain
from decayamplitude.combiner import ChainCombiner
from decayamplitude.resonance import Resonance
from decayamplitude.rotation import QN

from decayangle.decay_topology import Topology, Node
from decayangle.config import config as decayangle_config

from decayangle.lorentz import LorentzTrafo

from decayamplitude.rotation import wigner_capital_d

from collections import defaultdict

def constant_lineshape(*args):
    return 1

def make_four_vectors(phi_rf, theta_rf, psi_rf):
    from decayamplitude.backend import numpy as np

    # Make sure, the sorting is turned off

    # Given values
    # Lc -> p K pi
    m0 = 6.32397
    m12 = 9.55283383**0.5
    m23 = 26.57159046**0.5
    m13 = 17.86811729**0.5
    m1, m2, m3 = 1, 2, 3
    # Squared masses
    m0sq, m1sq, m2sq, m3sq, m12sq, m23sq = [x**2 for x in [m0, m1, m2, m3, m12, m23]]

    # Källén function
    def Kallen(x, y, z):
        return x**2 + y**2 + z**2 - 2 * (x * y + x * z + y * z)

    # Calculating missing mass squared using momentum conservation
    m31sq = m0sq + m1sq + m2sq + m3sq - m12sq - m23sq

    # Momenta magnitudes
    p1a = np.sqrt(Kallen(m23sq, m1sq, m0sq)) / (2 * m0)
    p2a = np.sqrt(Kallen(m31sq, m2sq, m0sq)) / (2 * m0)
    p3a = np.sqrt(Kallen(m12sq, m3sq, m0sq)) / (2 * m0)

    # Directions and components
    cos_zeta_12_for0_numerator = (m0sq + m1sq - m23sq) * (
        m0sq + m2sq - m31sq
    ) - 2 * m0sq * (m12sq - m1sq - m2sq)
    cos_zeta_12_for0_denominator = np.sqrt(Kallen(m0sq, m2sq, m31sq)) * np.sqrt(
        Kallen(m0sq, m23sq, m1sq)
    )
    cos_zeta_12_for0 = cos_zeta_12_for0_numerator / cos_zeta_12_for0_denominator

    p1z = -p1a
    p2z = -p2a * cos_zeta_12_for0
    p2x = np.sqrt(p2a**2 - p2z**2)
    p3z = -p2z - p1z
    p3x = -p2x

    # Energy calculations based on the relativistic energy-momentum relation
    E1 = np.sqrt(p1z**2 + m1sq)
    E2 = np.sqrt(p2z**2 + p2x**2 + m2sq)
    E3 = np.sqrt(p3z**2 + p3x**2 + m3sq)

    # Vectors
    p1 = np.array([0, 0, p1z, E1])
    p2 = np.array([p2x, 0, p2z, E2])
    p3 = np.array([p3x, 0, p3z, E3])

    # Lorentz transformation
    momenta = {i: p for i, p in zip([1, 2, 3], [p1, p2, p3])}
    tree1 = Topology(root=0, decay_topology=((2, 3), 1))

    # momenta = Topology(root=0, decay_topology=((1, 2), 3)).align_with_daughter(momenta, 3)
    # momenta = tree1.root.transform(LorentzTrafo(0, 0, 0, 0, -np.pi, 0), momenta)
    rotation = LorentzTrafo(0, 0, 0, phi_rf, theta_rf, psi_rf)

    momenta_23_rotated = tree1.root.transform(rotation, momenta)
    return momenta_23_rotated

def resonances():
    resonances1 = {
    (2,3): Resonance(Node((2, 3)), 0, -1, lineshape=constant_lineshape, argnames=[]),
    0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[])
    }

    resonances2 = {
        (1, 2): Resonance(Node((1, 2)), 3, -1, lineshape=constant_lineshape, argnames=[]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[])
    }

    resonances3 = {
        (1, 2): Resonance(Node((1, 2)), 1, -1, lineshape=constant_lineshape, argnames=[]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[])
    }

    resonances_dpd = {
        (2,3): Resonance(Node((2, 3)), 4, -1, lineshape=constant_lineshape, argnames=[]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[])
    }
    return resonances1, resonances2, resonances3, resonances_dpd

def test_threebody_1():
    decayangle_config.sorting = "off" 
    topology1 = Topology(
        0,
        decay_topology=((2,3), 1)
    )

    topology2 = Topology(
        0,
        decay_topology=((1, 2), 3)
    )

    resonances1, resonances2, resonances3, resonances_dpd = resonances()

    from decayamplitude.backend import numpy as np
    # momenta = make_four_vectors(np.linspace(0,np.pi,10), np.linspace(0,np.pi,10), np.linspace(0,np.pi,10))
    momenta = make_four_vectors(0.3, np.arccos(0.4), 0.5)


    final_state_qn = {
            1: QN(1, 1),
            2: QN(2, 1),
            3: QN(0, 1)
        }
    decay = DecayChain(
        topology = topology1,
        resonances = resonances1,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    decay2 = DecayChain(
        topology = topology2,
        resonances = resonances2,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    decay3 = DecayChain(
        topology = topology2,
        resonances = resonances3,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    decay_dpd = DecayChain(
        topology = topology1,
        resonances = resonances_dpd,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    decay_dpd_m = DecayChain(
        topology = topology1,
        resonances = resonances_dpd,
        momenta = momenta,
        final_state_qn = final_state_qn,
        convention = "minus_phi"
    )

    arguments1 = {
        resonances1[(2,3)].id : {
            "couplings":{
                (2, 2) : 1
            }
        }, 
        resonances1[0].id : {
            "couplings":{
                (2, 1) : 1
            }
        }
    }
    arguments2 = {
        resonances2[(1,2)].id : {
            "couplings":{
                (2, 3) : 1
            }
        }, 
        resonances2[0].id : {
            "couplings":{
                (2, 3) : 1
            }
        }
    }

    arguments3 = {
        resonances3[(1,2)].id : {
            "couplings":{
                (2, 3) : 1
            }
        }, 
        resonances3[0].id : {
            "couplings":{
                (2, 1) : 1
            }
        }
    }

    arguments_dpd = {
        resonances_dpd[(2,3)].id : {
            "couplings": {
                (4, 2): 1
            },
        },
        resonances_dpd[0].id : {
            "couplings": {
                (4, 3): 1
            }
        }
    }


    dpd_value = decay_dpd.matrix(-1, arguments_dpd)[(1, 2,0)]
    dpd_value_m = decay_dpd_m.matrix(-1, arguments_dpd)[(1, 2,0)]

    aligned_decay3 = AlignedChain(
        topology = topology2,
        resonances = resonances3,
        momenta = momenta,
        final_state_qn = final_state_qn,
        reference=decay_dpd,
        convention="helicity"
    )

    aligned_decay3_m = AlignedChain(
        topology = topology2,
        resonances = resonances3,
        momenta = momenta,
        final_state_qn = final_state_qn,
        reference=decay_dpd_m,
        convention="minus_phi"
    )
    value3 = aligned_decay3.aligned_matrix(-1, arguments3)[(1, 2, 0)]
    value3_m = aligned_decay3_m.aligned_matrix(-1, arguments3)[(1, 2, 0)]
    # this is a reference value copied from the output of the decayangle code
    # We can use this to harden against mistakes in the decayamplitude code
    assert np.allclose(dpd_value, (-0.14315554700441074 + 0.12414558894503328j))
    assert np.allclose(
        dpd_value_m, -0.03883258888101088 + 0.1854660829732478j
    )
    assert np.allclose(
        value3, -0.49899891547281655 + 0.030820810874496913j
    )
    assert np.allclose(
        value3_m, -0.37859261634645197 + 0.32652330831650717j
    )



    # assert np.allclose(
    #     terms_2_m[(-1, 1, 2, 0)][-1], -0.37859261634645197 + 0.32652330831650717j
    # )
    # assert np.allclose(unpolarized(full_amp), unpolarized(full_amp)[0])

def testShortThreeBodyAmplitude():
    """
    We can also combine the chains in a shorter way, by using the MultiChain class
    """
    final_state_qn = {
            1: QN(1, 1),
            2: QN(2, 1),
            3: QN(0, 1)
        }
    resonances1, resonances2, resonances3, resonances_dpd = resonances()
    momenta = make_four_vectors(1,2,np.linspace(0,np.pi,10))
    topology1 = Topology(
        0,
        decay_topology=((2,3), 1)
    )
    topology2 = Topology(
        0,
        decay_topology=((1, 2), 3)
    )

    chain1 = MultiChain.from_chains([
        DecayChain(
            topology = topology1,
            resonances = resonances1,
            momenta = momenta,
            final_state_qn = final_state_qn
        ),
        DecayChain(
            topology = topology1,
            resonances = resonances_dpd,
            momenta = momenta,
            final_state_qn = final_state_qn
        )
    ])

    chain2 = MultiChain.from_chains([
        DecayChain(
            topology = topology2,
            resonances = resonances2,
            momenta = momenta,
            final_state_qn = final_state_qn,
        ),
        DecayChain(
            topology = topology2,
            resonances = resonances3,
            momenta = momenta,
            final_state_qn = final_state_qn
        )
    ])

    merged_resonances = defaultdict(list)
    for key, resonance in resonances3.items():
        merged_resonances[key].append(resonance)
    for key, resonance in resonances2.items():
        merged_resonances[key].append(resonance)
    merged_resonances = dict(merged_resonances)

    merged_resonances[0] = [merged_resonances[0][0]]

    full = ChainCombiner([chain1, chain2])
    arguments = full.generate_couplings()

    matrix1 = full.combined_matrix(-1, arguments)
    matrix2 = full.combined_matrix(1, arguments)

    unpolarized, argnames = full.unpolarized_amplitude(full.generate_couplings(), complex_couplings=False)
    assert np.allclose(sum(abs(v)**2 for v in matrix1.values()) + sum(abs(v)**2 for v in matrix2.values()), unpolarized(*([1] * len(argnames))))
    assert np.allclose(unpolarized(*([1] * len(argnames))), unpolarized(*([1] * len(argnames)))[0])

    full2 = ChainCombiner([chain1, MultiChain(topology2, momenta=momenta, resonances=merged_resonances, final_state_qn=final_state_qn)])
  
    unpolarized2, argnames2 = full2.unpolarized_amplitude(full2.generate_couplings(), complex_couplings=False)

    # initial orientation should not affect the result
    assert np.allclose(unpolarized2(*([1] * len(argnames))), unpolarized2(*([1] * len(argnames2)))[0])

    assert np.allclose(unpolarized2(*([1] * len(argnames))) , unpolarized(*([1] * len(argnames))))


if __name__ == "__main__":
    testShortThreeBodyAmplitude()
    test_threebody_1()


from __future__ import annotations

from decayangle.decay_topology import Node, HelicityAngles, Topology
from decayamplitude.rotation import QN, Angular, clebsch_gordan, wigner_capital_d, convert_angular
from typing import Sequence, Union, Callable, Literal, Generator
from collections import namedtuple
from functools import cached_property
from decayamplitude.utils import sanitize

import warnings


LSTuple = namedtuple("LSTuple", ["l", "s"])
HelicityTuple = namedtuple("HelicityTuple", ["h1", "h2"])

class Resonance:
    __instances = {}
    __named_instances = {}
    __parameter_names = {}

    @classmethod
    def get_instance(cls, id:int) -> "Resonance":
        return cls.__instances[id]

    def __init__(self, node:Node, spin:Union[Angular, int] = None, parity:int = None, quantum_numbers:QN = None, lineshape = None, argnames = None, name = None, preserve_partity=True, scheme:Literal["ls", "helicity"]="ls") -> None:
        self.node = node
        self.preserve_partity = preserve_partity
        if quantum_numbers is None:
            if spin is None or parity is None:
                raise ValueError("Either quantum numbers or spin and parity must be provided")
            self.quantum_numbers = QN(spin, parity)
        else:
            self.quantum_numbers = quantum_numbers
        self.__daughters = None
        self.__lineshape = None
        if name is not None:
            self.__name = name
        else: 
            # we explicitly want to differentiate resonances with and without name
            self.__name = None
        self.__id = Resonance.register(self)

        if lineshape is not None:
            if argnames is None:
                raise ValueError("If a lineshape is provided, the argument names must be provided as well")
            self.register_lineshape(lineshape, argnames)
            self.__lineshape = lineshape
            self.__parameter_names = argnames
        if scheme not in ["ls", "helicity"]:
            raise ValueError("scheme must be either 'ls' or 'helicity'")
        self.__scheme = scheme

    @property
    def parameter_names(self) -> list[str]:
        return self.__parameter_names
                
    def argument_list(self, arguments:dict) -> list:
        return [arguments[name] for name in self.__parameter_names]

    @property
    def scheme(self) -> str:
        return self.__scheme

    @property
    def name(self) -> str:
        if self.__name is None:
            return f"ID_{self.id}"
        return self.__name
    
    @cached_property
    def sanitized_name(self) -> str:
        """
        Name sanitized for use in python code

        Returns:
            str: sanitized name
        """
        return sanitize(self.name)
    
    @property
    def descriptor(self) -> str:
        """
        Returns a string that describes the resonance
        """
        return f"{self.sanitized_name}_to_{'_'.join([d.sanitized_name for d in self.daughters])}"

    @property
    def id(self) -> int:
        return self.__id
    
    @id.setter
    def id(self, value:int):
        raise ValueError("The id of a resonance cannot be changed")

    @classmethod
    def register(cls, obj) -> int:
        instance_id = len(cls.__instances)
        cls.__instances[instance_id] = obj
        if obj.__name is not None:
            if not obj.__name in cls.__named_instances:
                # We have multiple resonances with the same name, we assume that they share a parameter set 
                # and thus we only need to register the name once
                # The parameter set is only for the lineshape and not for the couplings
                cls.__named_instances[obj.__name] = [obj]
            else:
                cls.__named_instances[obj.__name].append(obj)

        return instance_id
    
    def copy(self):
        return Resonance(self.node, quantum_numbers=self.quantum_numbers, lineshape=self.__lineshape, argnames=self.__parameter_names, name=self.__name, preserve_partity=self.preserve_partity, scheme=self.__scheme)

    @property
    def lineshape(self) -> Callable:
        return self.__lineshape

    @property
    def tuple(self) -> tuple:
        return self.node.value
    
    @property
    def daughter_qn(self) -> list[QN]:
        if self.daughters is None:
            raise ValueError(f"{self.node}: Daughter quantum numbers not set! This should happen in the DecayChainNode class. The resonance class should only be used as part of a DecayChain!")
        return [daughter.quantum_numbers for daughter in self.__daughters]
    

    @property   
    def daughters(self) ->  list["Resonance" | Node]:
        if self.__daughters is None:
            raise ValueError(f"{self.node}: Daughters not set! This should happen in the DecayChainNode class. The resonance class should only be used as part of a DecayChain!")
        return self.__daughters

    @daughters.setter
    def daughters(self, daughters: list[Resonance | Node]):
        self.__daughters = daughters
    
    def __str__(self):
        return f"Resonance {self.name} at {self.node} with {self.quantum_numbers}"
    
    def __repr__(self):
        return self.__str__()
    
    @convert_angular
    def helicity_from_ls(self, h0:Union[Angular, int], h1:Union[Angular, int], h2:Union[Angular, int], couplings:dict[LSTuple, float], arguments:dict):
        """
        This function translates from the ls basis into the helicity basis.
        The linehspae funcitons can depend on L and S.
        It is important, that particle 2 convention from Jacob-Wick is used!
        Otherwise the direct mapping of LS to helicity does not work.

        Note:
            Wrapper convert_angular ensures, that inside the function only integes arrive as values for h.

        arguments:
        h0: int
            Helicity of the resonance
        h1: int
            Helicity of the first daughter
        h2: int
            Helicity of the second daughter
        couplings: dict
            The couplings of the resonance.
            Format is {(l, s): value}
        arguments: dict
            The arguments for the lineshape function. The keys are the names of the arguments.
        
        """
        q1, q2 = self.daughter_qn
        j1, j2 = q1.angular.value2, q2.angular.value2

        return sum(
            coupling * 
            self.lineshape(l,s, *self.argument_list(arguments)) * 
            (l + 1) ** 0.5 /
            (self.quantum_numbers.angular.value2 + 1) ** 0.5 *
            clebsch_gordan(j1, h1, j2, -h2, s, h1- h2) *
            clebsch_gordan(l, 0, s, h1 - h2, self.quantum_numbers.angular.value2, h1 - h2)
            for (l, s), coupling in couplings.items()
        )


    def __construct_couplings(self, arguments:dict) -> dict[LSTuple, float]:
        """
        TEMPORARY VERSION FINAL SOLUTION NOT YET CLEAR

        We constuct the couplings for the resonance from real numbers found in the arguments dict
        Thus we need to know, which arguments belong to the couplings of the resonance
        The arguments are a dict of the form {id: {parameter_name: value}} where the id is the id of the resonance
        The ls couplings are a dict of the form {(l,s): value_r } under the name ls_couplings
        """
        couplings = arguments[self.id]["couplings"]
        return {LSTuple(*key): value for key, value in couplings.items()}

    def generate_couplings(self, conserve_parity:bool = True) -> dict[Union[LSTuple, HelicityTuple], float]:
        """
        This function should be used to generate the couplings for the resonance
        """

        if self.scheme == "helicity":
            return {
                "couplings": {
                    HelicityTuple(h1, h2): 1
                    for h1 in self.daughter_qn[0].projections(return_int=True)
                    for h2 in self.daughter_qn[1].projections(return_int=True)
                }
            }
        qn1, qn2 = self.daughter_qn
        qn0 = self.quantum_numbers

        if qn0.angular.value2 % 2 != (qn1.angular.value2 + qn2.angular.value2) % 2:
            raise ValueError(f"Angular momentum of resonance {self.name} at {self.node} is not compatible with the angular momenta of the daughters {qn1}, {qn2}.")
        possible_states = set(QN.generate_L_states(qn0, qn1, qn2))
        if not conserve_parity:
            qn0_bar = QN(qn0.angular.angular_momentum, -qn0.parity)
            possible_states = possible_states.union(set(QN.generate_L_states(qn0_bar, qn1, qn2)))
        if len(possible_states) == 0:
            raise ValueError(f"No possible states for resonance {self.name} at {self.node} with {qn0}, {qn1}, {qn2}.")
        return {
                "couplings": {
                    LSTuple(l.value2, s.value2): 1
                    for l, s in possible_states
                }
            }
    
    def direct_helicity_coupling(self, arguments, h1, h2):
        return arguments[self.id]["couplings"][(h1, h2)] * self.lineshape(h1, h2, *self.argument_list(arguments))
    
    @convert_angular
    def amplitude(self, h0:Union[Angular, int], h1:Union[Angular, int], h2:Union[Angular, int], arguments:dict):
        if self.scheme == "ls":
            couplings = self.__construct_couplings(arguments)
        # lineshape_args = self.__construct_arguments(arguments)
            coupling = self.helicity_from_ls(h0, h1, h2, couplings ,arguments)
        elif self.scheme == "helicity":
            coupling = self.direct_helicity_coupling(arguments, h1, h2)
        else:
            raise ValueError(f"Scheme must be either 'ls' or 'helicity' but is {self.scheme}")
        # particle 2 convention from Jacob-Wick is used!
        j2 = self.daughter_qn[1].angular.value2
        return coupling * (-1) ** ((j2 - h2) / 2) 
    
    def register_lineshape(self, lineshape_function:Callable, parameter_names: list[str]):
        if self.__lineshape is not None:
            raise ValueError("Lineshape already set")
        self.__lineshape = lineshape_function
        self.__parameter_names = {}
        for parameter_name in parameter_names:
            if parameter_name not in self.__parameter_names:
                type(self).__parameter_names[parameter_name] = self
        return self
    
def flat_generator(tpl: tuple) -> Generator[Union[tuple, int]]:
    """
    Generator that flattens a tuple of tuples into a single tuple.
    """
    if isinstance(tpl, tuple) and not isinstance(tpl, list):
        for item in tpl:
            yield from flat_generator(item)
    elif isinstance(tpl, int):
        yield tpl
    else:
        raise TypeError(f"flat_generator: Expected tuple or int, got {type(tpl)} with value {tpl}.")

class ResonanceDict:
    def __init__(self, resonances: dict[tuple, Resonance | list[Resonance]]) -> None:
        self.resonances = resonances

        self.stable_value_dict = {self.stable_value(key): value for key, value in resonances.items()}
        if len(self.stable_value_dict) != len(resonances):
            duplicate_keys = set()
            stable_keys = set(self.stable_value_dict.keys())
            for key in resonances:
                stable_key = self.stable_value(key)
                if stable_key not in stable_keys:
                    duplicate_keys.add(key)
                else:
                    stable_keys.remove(stable_key)
            raise ValueError(f"ResonanceDict: Duplicate keys found in the resonance dictionary: {duplicate_keys}. This is not allowed.")
        warnings.warn(f"ResonanceDict initialized with {len(self.stable_value_dict)} resonances.")

    @classmethod
    def stable_value(cls, value:tuple) -> tuple[int] | int:
        """
        Returns a stable value for the resonance dictionary keys.
        This is needed, to make sure all nodes are treated the same.
        """
        base_value = tuple(sorted(flat_generator(value)))
        if len(base_value) == 1:
            return base_value[0]
        return base_value

    def __getitem__(self, key:tuple) -> Resonance:
        """
        Returns the resonance for the given key.
        The key is a tuple of the form (l, s, h1, h2) or (h1, h2) depending on the scheme.
        """
        stable_key = self.stable_value(key)
        return self.stable_value_dict[stable_key]
    
    def __contains__(self, key:tuple) -> bool:
        """
        Returns True if the resonance dictionary contains the key.
        """
        stable_key = self.stable_value(key)
        return stable_key in self.stable_value_dict
    
    def __iter__(self):
        """
        Returns an iterator over the resonances in the dictionary.
        """
        return iter(self.stable_value_dict.keys())

    def __len__(self):
        """
        Returns the number of resonances in the dictionary.
        """
        return len(self.stable_value_dict)
    
    def __str__(self):
        """
        Returns a string representation of the resonance dictionary.
        """
        return f"ResonanceDict with {len(self)} resonances: {', '.join([str(key) for key in self.stable_value_dict.keys()])}"
    
    def __repr__(self):
        """
        Returns a string representation of the resonance dictionary.
        """
        return self.__str__()
    
    def __eq__(self, other):
        """
        Returns True if the resonance dictionaries are equal.
        """
        if not isinstance(other, ResonanceDict):
            return False
        return self.stable_value_dict == other.stable_value_dict
    
    def __ne__(self, other):
        """
        Returns True if the resonance dictionaries are not equal.
        """
        return not self.__eq__(other)  

    def get(self, key:tuple, default: Resonance| None=None) -> Resonance | list[Resonance] | None:
        """
        Returns the resonance for the given key or the default value if the key is not found.
        """
        stable_key = self.stable_value(key)
        return self.stable_value_dict.get(stable_key, default)

    def items(self):
        """
        Returns an iterator over the items in the resonance dictionary.
        """
        return self.stable_value_dict.items()
    
    def values(self):
        """
        Returns an iterator over the values in the resonance dictionary.
        """
        return self.stable_value_dict.values()
    
    def keys(self):
        """
        Returns an iterator over the keys in the resonance dictionary.
        """
        return self.stable_value_dict.keys()
    
    def filter_by_topology(self, topo: Topology) -> ResonanceDict:
        """
        Returns a list of resonances that are relevant for the given topology.
        """
        filtered_resonances ={
            self.stable_value(node.value): self[node.value] for node in topo.nodes.values() if not node.final_state
        } 
        filtered_resonances =  ResonanceDict(filtered_resonances)
        if filtered_resonances.get(
            self.stable_value(topo.root.value), default=None
        ) is None:
            raise ValueError(f"ResonanceDict: No root resonance found for topology {topo}. This is not allowed.")
        return filtered_resonances

    def __set_item__(self, key:tuple, value:Resonance | list[Resonance]):
        """
        Sets the resonance for the given key.
        The key is a tuple of the form (l, s, h1, h2) or (h1, h2) depending on the scheme.
        """
        stable_key = self.stable_value(key)
        self.stable_value_dict[stable_key] = value


from typing import Union, Optional, Callable, Literal
from itertools import product
import warnings as warnings
from functools import cached_property

from decayangle.decay_topology import Topology, Node

from decayangle.decay_topology import Topology, HelicityAngles, WignerAngles
from decayamplitude.particle import Particle
from decayamplitude.resonance import Resonance
from decayamplitude.rotation import QN, wigner_capital_d, Angular, convert_angular
from decayamplitude.resonance import ResonanceDict
from decayamplitude.backend import numpy as np

from decayamplitude.utils import _create_function, sanitize

class DecayChainNode:
    """
    Class to represent a node in the decay chain. This utilizes the Node class from decayangle. 
    A Node has a resonance, and a topology, to make senese of its position in the decay chain. The node value only has a meaning in the context of the topology.
    """


    def __init__(self, node:Node, resonances: dict[tuple, Resonance] | ResonanceDict, final_state_qn: dict[int, QN | Particle], topology:Topology, convention:Literal["helicity", "minus_phi"]="helicity") -> None:
        """
        Initializes a DecayChainNode object. The object will contain a resonance and a topology.

        Parameters:
        node: Node
            The node of the decay chain. This is a node of the decay topology as defined in `decayangle`
        resonances: dict[tuple, Resonance]
            A dictionary with the resonances of the decay chain. The keys are the tuples of the nodes, the values are the resonances
        final_state_qn: dict[tuple, QN]
            A dictionary with the quantum numbers of the final state particles. The keys are the tuples of the nodes, the values are the quantum numbers
        topology: Topology
            The topology of the decay chain. This is a topology as defined in `decayangle`
        convention: str
            The convention of the decay chain. This is either "helicity" or "minus_phi". The default is "helicity"
        """
        # this check needs to happen first to avoid errors

        if node.value not in topology.nodes:
            # someone may have initialized the root with a tuple instead of 0
            if node.tuple == topology.root.tuple:
                self.node = topology.root
            else:
                raise ValueError(f"Node {node} not in topology {topology}")
        else:
            self.node = topology.nodes[node.value]
        
        self.tuple = self.node.tuple
        self.__is_root = self.node == topology.root
            
        if not isinstance(resonances, ResonanceDict):
            # if the resonances are not a ResonanceDict, we convert them to one
            resonances = ResonanceDict(resonances)
        self.resonance: Union[Resonance, None] = resonances.get(self.tuple, resonances.get(self.node.value, None))
        if self.resonance is None and not self.final_state:
            raise ValueError(f"Resonance for {self.node.tuple} not found. Every internal node must have a resonance to describe its behaviour!")

        self.resonances = resonances
        self.topology = topology
        self.final_state_qn = final_state_qn
        self.convention = convention
            
        self.daughters = [
                    DecayChainNode(daughter, resonances, self.final_state_qn, topology, convention=self.convention)
                    for daughter in self.node.daughters
            ]
        
        if not self.final_state:
            if self.resonance is None:
                raise ValueError(f"Resonance for {self.tuple} not found. Every internal node must have a resonance to describe its behaviour!")
            # set the daughters of the resonance
            self.quantum_numbers = self.resonance.quantum_numbers
            self.resonance.daughters = [daughter for daughter in self.daughters]
        else:
            self.quantum_numbers = self.final_state_qn[self.tuple]

    @property
    def final_state(self):
        """
        Returns:
        bool
            True if the node is a final state particle, False otherwise
        """
        return self.node.final_state
    
    @property
    def name(self) -> str:
        """
        Returns:
        str
            The name of the node
        """
        if self.resonance is not None:
            return self.resonance.name
        if isinstance(self.quantum_numbers, Particle):
            if self.quantum_numbers.name is not None:
                return f"{sanitize(self.quantum_numbers.name)}"
            return f"particle_type_{self.quantum_numbers.type_id}"
        return f"particle_{self.node.value}"
    
    @property
    def sanitized_name(self) -> str:
        """
        Returns:
        str
            The sanitized name of the node
        """
        if self.resonance is not None:
            return self.resonance.sanitized_name
        return sanitize(self.name)
    
    @property
    def is_root(self):
        """
        Returns:
        bool
            True if the node is the root of the decay chain, False otherwise
        """
        return self.__is_root

    @property
    def quantum_numbers(self) -> QN:
        """
        Returns:
        QN
            The quantum numbers of the node
        """
        return self.__qn 
    
    @property
    def decay_tuple(self) -> tuple:
        """
        Returns:
        tuple
            The decay tuple of the node. This is the tuple of the nodes daughters values.
        """
        return tuple([daughter.node.value for daughter in self.daughters])
    
    @quantum_numbers.setter
    def quantum_numbers(self, qn: QN):
        """
        Sets the quantum numbers of the node. This is used to set the quantum numbers of the resonance.
        """
        self.__qn = qn

    def __helicity_angles(self, angles:HelicityAngles) -> tuple:
        """
        Returns the helicity angles of the node. This is used to calculate the amplitude of the decay chain.

        Parameters:
            angles: HelicityAngles
                The helicity angles of the node. This is used to calculate the amplitude of the decay chain. Helicity angles are defined in the `decayangle` library.
        Returns:
            tuple
                The helicity angles of the node. This is used to calculate the amplitude of the decay chain.
        """
        if self.convention == "helicity":
            return (angles.phi_rf, angles.theta_rf, 0)
        if self.convention == "minus_phi":
            return (angles.phi_rf, angles.theta_rf, -angles.phi_rf)
        raise ValueError(f"Convention {self.convention} not known")

    @convert_angular
    def amplitude(self, h0:Union[Angular, int], lambdas:dict, helicity_angles:dict[tuple,HelicityAngles], arguments:dict):
        """
        The amplitude of a single node given the helicity of the decaying particle
        The helicities of the daughters will be generated from here, recursively
        the arguments are the couplings of the resonances and are a global dict, which will be passed through the recursion
        This means, that all arguments for lineshapes will be provided positional 

        parameters:
        h0: int
            The helicity of the decaying particle
        lambdas: dict
            The helicites of the mother and the final state particles
        arguments: dict
            The couplings of the resonances and the resonance parameters
        
        returns:
        float
            The amplitude of the decay as a generator

        """
        if self.final_state:
            # we do not have a resonance and we have no daughters
            yield 1.
        else:
            d1, d2 = self.daughters
            if d1.final_state:
                d1_helicities = [lambdas[d1.tuple]]
            else:
                d1_helicities = d1.quantum_numbers.projections(return_int=True)
            if d2.final_state:
                d2_helicities = [lambdas[d2.tuple]]
            else:
                d2_helicities = d2.quantum_numbers.projections(return_int=True)

            for h1 in d1_helicities:
                for h2 in d2_helicities:
                    for A_1 in d1.amplitude(h1, lambdas, helicity_angles, arguments):
                        for A_2 in d2.amplitude(h2, lambdas, helicity_angles, arguments):
                            # TODO: add explicit handling of the arguments for the lineshape
                            A_self = self.resonance.amplitude(h0, h1, h2, arguments) * np.conj(wigner_capital_d(*self.__helicity_angles(helicity_angles[self.decay_tuple]), self.quantum_numbers.angular.value2, h0, h1 - h2))
                            yield A_1 * A_2 * A_self * (self.quantum_numbers.angular.value2 + 1)**0.5

class DecayChain:
    """
    Class to represent a decay chain. This is a topology in connection with a set of resonances. One resonance for each internal node in the topology.
    """

    def __init__(self, topology:Topology, resonances: dict[tuple, Resonance] | ResonanceDict, momenta: dict, final_state_qn: dict[int, QN | Particle], convention:Literal["helicity", "minus_phi"]="helicity") -> None:
        self.topology = topology
        if not isinstance(resonances, ResonanceDict):
            # if the resonances are not a ResonanceDict, we convert them to one
            resonances = ResonanceDict(resonances)
        self.resonances = resonances
        self.momenta = momenta
        self.final_state_qn = final_state_qn
        self.convention = convention

        self.root_resonance = self.resonances.get(self.topology.root.value)
        if self.root_resonance is None:
            self.root_resonance = self.resonances.get(self.topology.root.tuple)
        if self.root_resonance is None:
            raise ValueError(f"No root resonance found! The root resonance should be the decaying particle. The lineshape is irrelelevant for decay processes, but the quantum numbers are crucial! Define a root resonance under the key {self.topology.root.value} or {self.topology.root.tuple}. Available resonances: {list(self.resonances.keys())}")

        # we need a sorted version of the particle keys to map matrix elements to the correct particle helicities later
        self.final_state_keys = sorted(final_state_qn.keys())
        helicities = Angular.generate_helicities(*[final_state_qn[key] for key in self.final_state_keys])
        self.helicities = [
            {key: helicity[i] for i, key in enumerate(self.final_state_keys)}
            for helicity in helicities
        ]
        self.helicity_tuples = helicities
        self.resonance_list = list(resonances.values())
    
    @property
    def nodes(self):
        return list(
            DecayChainNode(node, self.resonances, self.final_state_qn, self.topology, self.convention)
            for node in self.topology.nodes.values()
        )
    
    @property
    def final_state_nodes(self) -> list[DecayChainNode]:
        return [node for node in self.nodes if node.final_state]

    @cached_property
    def helicity_angles(self):
        return self.topology.helicity_angles(momenta=self.momenta, convention=self.convention)

    @property
    def root(self):
        return DecayChainNode(self.topology.root, self.resonances, self.final_state_qn, self.topology, self.convention)

    @property
    def chain_function(self):

        def f(h0, lambdas:dict, arguments:dict):
            amplitudes = [
                 amplitude
                for amplitude in self.root.amplitude(h0, lambdas, self.helicity_angles, arguments)
            ]
            prefactor = 1/(self.root.resonance.quantum_numbers.angular.value2 + 1)**0.5
            return prefactor * sum(
               amplitudes
            )

        return f
    
    @property
    def matrix(self):
        """
        Returns a function, which will not produce the chain function for a single set of helicities, but will rather return a matrix with all possible helicities. 
        The matrix will be an actual matrix and not a dict, since we want to use it later to perform matrix operations.
        """

        f = self.chain_function
        def matrix(h0, arguments:dict):
            return {
                tuple([lambdas[key] for key in self.final_state_keys]): f(h0, lambdas, arguments)
                for lambdas in self.helicities
            }
        
        return matrix
    
    def generate_couplings(self):
        """
        Returns all LS couplings for the decay chain
        """
        return {
            node.resonance.id: node.resonance.generate_couplings(node.resonance.preserve_partity)
            for node in self.nodes
            if not node.final_state
        }
    
    @property
    def resonance_params(self):
        resonances = [resonance for resonance in self.resonance_list]
        resonance_parameter_names = [name for resonance in resonances for name in resonance.parameter_names]

        if len(set(resonance_parameter_names)) != len(resonance_parameter_names):
            from collections import Counter
            c = Counter(resonance_parameter_names)
            # raise ValueError(f"Parameter names are not unique: {', '.join([name for name, count in c.items() if count > 1])}")
        return list(set(resonance_parameter_names))

class AlignedChain(DecayChain):
    """
    The aligned version of the decay chain. This is used to calculate the aligned amplitude, which is the amplitude in the final state helicity frame as defined by a reference topology or reference chain.
    """

    def __init__(self, topology:Topology, resonances: dict[tuple, Resonance], momenta: dict, final_state_qn: dict[int, QN | Particle], reference:Union[Topology, DecayChain], wigner_rotation: dict[tuple, WignerAngles]= None, convention:Literal["helicity", "minus_phi"]="helicity") -> None:
        """
        Initializes an AlignedChain object. The object will contain a list of DecayChain objects.
        Parameters:
        topology: Topology
            The topology of the decay chain
        resonances: dict[tuple, Resonance]
            A dictionary with the resonances of the decay chain. The keys are the tuples of the nodes, the values are the resonances
        momenta: dict
            The momenta of the decay chain
        final_state_qn: dict[tuple, QN]
            A dictionary with the quantum numbers of the final state particles
        reference: Topology
            The reference topology for the decay chain. This is used to calculate the alignment.
        wigner_rotation: dict[tuple, WignerAngles]
            A dictionary with the Wigner angles for the decay chain. This is used to calculate the alignment. Calculated if not provided.
        convention: str
            The convention of the decay chain. This is either "helicity" or "minus_phi". The default is "helicity"
        """
        self.reference: Topology = reference if isinstance(reference, Topology) else reference.topology
        self.topology = topology

        super().__init__(topology, resonances, momenta, final_state_qn, convention)
        if wigner_rotation is None:
            if isinstance(reference, DecayChain):
                if reference.convention != convention:
                    raise ValueError(f"Reference and chain must have the same convention. Found reference: {reference.convention} and self: {convention}!")
            self.wigner_rotation = self.reference.relative_wigner_angles(self.topology, momenta, convention=self.convention)
        else:
            self.wigner_rotation = wigner_rotation
        # we want the tuple versions of the helicities, since we use them as tuples
        self.wigner_dict = {
            key: {
                (h_, h): np.conj(wigner_capital_d(*self.wigner_rotation[key], final_state_qn[key].angular.value2, h, h_))
                for h in final_state_qn[key].angular.projections(return_int=True)
                for h_ in final_state_qn[key].angular.projections(return_int=True)
            }
            for key in self.final_state_keys
        }

    def to_tuple(self, lambdas:dict):
        return tuple([lambdas[key] for key in self.final_state_keys])

    @property
    def aligned_matrix(self):
        """
        Returns a function, which will return the amplitude for a given set of helicities. 
        The function will use the matrix to perform the calculation.
        """
        m = self.matrix
        def f(h0, arguments:dict):
            matrix = m(h0, arguments)
            aligned_matrix = {
                self.to_tuple(lambdas): sum(
                    matrix[self.to_tuple(lambdas_)]
                    * np.prod(np.array([
                            self.wigner_dict[key][(lambdas[key], lambdas_[key])] for key in self.final_state_keys
                        ]), 
                        axis=0)
                    for lambdas_ in self.helicities
                )
                for lambdas in self.helicities
            }
            return aligned_matrix
        
        return f
    
    def aligned_matrix_function(self, couplings:dict) -> Callable:
        """
        Returns a function, which will return the amplitude for a given set of helicities. 
        The function will use the matrix to perform the calculation.
        """
        aligned_matrix = self.aligned_matrix
        def f(arguments: dict):
            h0 = arguments.pop("h0")
            return aligned_matrix(h0, arguments)

        return _create_function(
            ["h0"] + self.resonance_params, couplings, f
        )

    
class MultiChain(DecayChain):
    @classmethod
    def create_chains(cls, resonances: dict[tuple, tuple[Resonance]] | ResonanceDict, topology: Topology ) -> list[dict[tuple, Resonance]]:
        """
        Creates all possible chains from a dictionary with lists of reonances for each isobar
        """
        if not isinstance(resonances, ResonanceDict):
            # if the resonances are not a ResonanceDict, we convert them to one
            resonances = ResonanceDict(resonances)

        if topology is None:
            raise ValueError("Topology must be provided to create chains from resonances")
        # with a given topology we can restrict the resonances to the nodes in the topology
        # this is usefull, if we only have one global dict of resonances
        filtered_resonances = resonances.filter_by_topology(topology)
        chains =list( product(*[filtered_resonances[key] for key in filtered_resonances.keys() ]) )
        return [
            ResonanceDict({
                key: chain[i].copy()
                for i, key in enumerate(filtered_resonances.keys())
            })
            for chain in chains
        ]

    @classmethod
    def from_chains(cls, chains: list[DecayChain]) -> "MultiChain":
        new_obj = cls(chains[0].topology, chains[0].momenta, chains[0].final_state_qn, chains=chains)
        if any(chain.topology != chains[0].topology for chain in chains):
            raise ValueError("All chains must have the same topology")
        return new_obj

    def __init__(self, topology:Topology, momenta: dict, final_state_qn: dict[int, QN | Particle], resonances: Optional[dict[tuple, tuple[Resonance]]]=None, chains: Optional[list[DecayChain]]=None, convention:Optional[Literal["minus_phi", "helicity"]]="helicity") -> None:
        """
        Initializes a MultiChain object. The object will contain a list of DecayChain objects.

        Parameters:
        topology: Topology
            The topology of the decay chain
        momenta: dict
            The momenta of the decay chain
        final_state_qn: dict[tuple, QN]
            The quantum numbers of the final state particles
        resonances: dict[tuple, tuple[Resonance]] | ResonanceDict
            A dictionary with a list of resonances for each isobar
        chains: list[DecayChain]
            A list of DecayChain objects
        convention: str
            The convention of the decay chain. Default is "helicity"
        """

        if chains is not None:
            self.chains = chains
            # I will stick with a default value for the convention. It will have no effect if chains are provided
            # if convention is not None:
            #     raise ValueError("Convention must not be set if chains are provided directly")
            if not all(chain.convention == chains[0].convention for chain in chains):
                raise ValueError("All chains must have the same convention")
            self.convention = chains[0].convention
        elif resonances is not None:
            if not isinstance(resonances, ResonanceDict):
                print(resonances.keys())
                # if the resonances are not a ResonanceDict, we convert them to one
                resonances = ResonanceDict(resonances)
            resonant_nodes = [node for node in topology.nodes.values() if not node.final_state]
            if any(node.value not in resonances and node.tuple not in resonances for node in resonant_nodes):
                warnings.warn(f"Not all nodes have a resonance assigned: {resonances.keys()}, {list(map(lambda x: x.value,resonant_nodes))}")
            self.chains = [
                DecayChain(topology, chain_definition, momenta, final_state_qn, convention)
                for chain_definition in type(self).create_chains(resonances, topology)
            ]
            def chain_filter(chain: DecayChain) -> bool:
                try:
                    chain.generate_couplings()
                except ValueError as e:
                    warnings.warn(f"Chain {chain} is not valid: {e}")
                    return False
                return True
            self.chains = [chain for chain in self.chains if chain_filter(chain)]
            if not self.chains:
                raise ValueError("There are no valid chains in the provided resonances! Check the resonances and the topology! Or check the quantum numbers!")
            self.convention = convention
        else:
            raise ValueError("Either resonances or chains must be provided")
        if chains is not None and resonances is not None:
            raise ValueError("Either resonances or chains must be provided")

    @property
    def chain_function(self) -> Callable:
        def f(h0, lambdas:dict, arguments:dict):
            return sum(
                chain.chain_function(h0, lambdas, arguments)
                for chain in self.chains
            )
        return f
    
    @property
    def resonance_list(self) -> list[Resonance]:
        return [
            resonance for chain in self.chains
            for resonance in chain.resonance_list
        ]
    
    @property
    def final_state_keys(self) -> list[Union[tuple, int]]:
        return self.chains[0].final_state_keys
    
    @property
    def topology(self):
        return self.chains[0].topology

    @property
    def matrix(self):
        def dict_sum(*dtcs):
            if len(dtcs) == 1:
                return dtcs[0]
            if len(dtcs) == 0:
                raise ValueError("No dicts to sum")
            if any(set(dtcs[0].keys()) != set(dtc.keys()) for dtc in dtcs):
                raise ValueError("Keys of the dicts do not match")
            return {
                key: sum(dtc[key] for dtc in dtcs)
                for key in dtcs[0].keys()
            }

        def matrix(h0, arguments:dict):
            return dict_sum(
                *[chain.matrix(h0, arguments)
                for chain in self.chains]
            )
        return matrix
    
    @property
    def root(self):
        return self.chains[0].root
    
    @property
    def nodes(self):
        return self.chains[0].nodes
    
    @property
    def helicities(self):
        return self.chains[0].helicities
    
    @property
    def helicity_tuples(self):
        return self.chains[0].helicity_tuples
    
    def generate_couplings(self):
        """
        Returns all LS couplings for the decay chain
        """
        coupling_dict = {}
        for chain in self.chains:
            coupling_dict.update(chain.generate_couplings())
        return coupling_dict
    
    @property
    def root_resonance(self) -> Union[Resonance, None]:
        if all(chain.root_resonance.quantum_numbers == self.chains[0].root_resonance.quantum_numbers for chain in self.chains):
            return self.chains[0].root_resonance
        return None
    
class AlignedMultiChain(MultiChain):
    @classmethod
    def from_chains(cls, chains: list[DecayChain], reference:Union[Topology, DecayChain]) -> "AlignedMultiChain":
        return cls(
            chains[0].topology,
            chains[0].momenta,
            chains[0].final_state_qn,
            reference,
            chains=chains
        )

    @classmethod
    def from_multichain(cls, multichain: MultiChain, reference:Union[Topology, DecayChain]) -> "AlignedMultiChain":
        return cls.from_chains(
            multichain.chains,
            reference
        )

    def __init__(self, topology:Topology, momenta: dict, final_state_qn: dict[int, QN | Particle], reference:Union[Topology, DecayChain], resonances: Optional[dict[tuple, tuple[Resonance]] | ResonanceDict] = None, chains: Optional[list[DecayChain]] = None, wigner_rotation: dict[tuple, WignerAngles]= None, convention:Literal["helicity", "minus_phi"]="helicity") -> None:
        super().__init__(topology, momenta, final_state_qn, resonances=resonances, chains=chains, convention=convention)
        self.reference: Topology = reference if isinstance(reference, Topology) else reference.topology
        if wigner_rotation is None:
            self.wigner_rotation = self.reference.relative_wigner_angles(self.topology, momenta, convention=self.convention)
        else:
            self.wigner_rotation = wigner_rotation

        self.wigner_dict = {
            key: {
                (h_, h): np.conj(wigner_capital_d(*self.wigner_rotation[key], final_state_qn[key].angular.value2, h, h_))
                for h in final_state_qn[key].angular.projections(return_int=True)
                for h_ in final_state_qn[key].angular.projections(return_int=True)
            }
            for key in self.final_state_keys
        }

    def to_tuple(self, lambdas:dict):
        return tuple([lambdas[key] for key in self.final_state_keys])
        

    @property
    def aligned_matrix(self):
        """
        Returns a function, which will return the amplitude for a given set of helicities. 
        The function will use the matrix to perform the calculation.
        """
        m = self.matrix
        def f(h0,  arguments:dict):
            matrix = m(h0, arguments)
            aligned_matrix = {
                self.to_tuple(lambdas): sum(
                    matrix[self.to_tuple(lambdas_)]
                    * np.prod(np.array([
                            self.wigner_dict[key][(lambdas[key], lambdas_[key])] for key in self.final_state_keys
                        ]), 
                        axis=0)
                    for lambdas_ in self.helicities
                )
                for lambdas in self.helicities
            }
            return aligned_matrix
        
        return f









# This file was generated by decaytreeedit-backend
# Copyright © 2025 Kai Habermann

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Author(s) shall be credited in any publication or presentation that uses this software, including but not limited to conference presentations, journal articles, and theses. The Authors shall be credited by reference to the decayangle and decayamplitude repositories, as well as the citation of the original paper: https://doi.org/10.1103/PhysRevD.111.056015

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 

from decayamplitude.resonance import Resonance
from decayamplitude.rotation import QN
from decayamplitude.chain import MultiChain
from decayamplitude.combiner import ChainCombiner
from decayamplitude.particle import Particle, DecaySetup
from decayamplitude.backend import numpy as np
from decayangle.decay_topology import Topology, Node
from decayangle.config import config as decayangle_config
from decayamplitude.kinematics_helpers import mass_from_node

import jax

decayangle_config.backend = "numpy"
decayangle_config.sorting = "value"  # this makes sure, that the order of entries in a node is handeled internaly and we do not have to worry about it

if hasattr(decayangle_config, "parallel_cores"):
    decayangle_config.parallel_cores = "auto"
    print("parallel_cores: ", decayangle_config.get_parallel_cores())
if hasattr(decayangle_config, "parallel_chunk_size"):
    decayangle_config.parallel_chunk_size = "auto"
    print("parallel_chunk_size: ", decayangle_config.get_parallel_chunk_size())

import os 
os.environ['XLA_FLAGS'] = "--xla_disable_hlo_passes=constant_folding"
os.environ["JAX_USE_SIMPLIFIED_JAXPR_CONSTANTS"] = "True"
jax.config.update("jax_enable_x64", True)

from jax import jit, grad

def constant_lineshape(L, S, *args):
    #This is a dummy lineshape function, it does nothing, but is needed for the code to work
    # It is suited to model a mother resonance for example
    return 1.0

def BW_lineshape(mass):
    # usually the function is called for a fixed set of points in the phasespace. 
    # Thus we can give the mass values as fixed
    # This allows the compiler to optimize the function better
    # decayamplitude will provide l and s values per default for the resonances, but not the mass values
    def BW(l,s, m0, gamma):
        return 1 / (mass**2 - m0**2 + 1j * mass * gamma) 
    return BW


topologies = [
        Topology(0, decay_topology=((2, 3), 1)),
        Topology(0, decay_topology=((1, 3), 2))
    ]
        
final_state_qn = {
    1: Particle(spin=1, parity=1, name="Lambda_c()+"), 
    2: Particle(spin=0, parity=-1, name="Dbar0"), 
    3: Particle(spin=0, parity=-1, name="K-"), 
}
setup = DecaySetup(final_state_particles=final_state_qn)
topologies = [
    setup.symmetrize(topology) for topology in topologies
]

def amplitude(momenta):
    # momenta has the form of 
    # {
    #  1: np.array(shape=(...,4)), with indices 0,1,2,3 for p_x,p_y,p_z,E
    #  2: np.array(shape=(...,4)), with indices 0,1,2,3 for p_x,p_y,p_z,E
    #  ...
    #}

    resonances = {
        0:    [
            Resonance(Node(0), quantum_numbers=QN(1, 1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name='Lambda_b()0') 
            # Mother particle, set the QN to what they need to be for your decay. Usually a weak initial decay is assumed, so parity is not preserved. If this is not the case set it to True here or in the frontend.
                    # Replace the lineshape with a function of your liking, the call signatrue has to be f(L, S, *args), where the args will be exposed under the names you give in argnames
                    # The mass points at which to evaluate the lineshape are not given at runtime. Look at the bw_lineshpe for details
            ],
        
        (2, 3): [
        Resonance(Node((2, 3)), quantum_numbers=QN(2, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=True, name='D_s_Kmatrix'),
            Resonance(Node((2, 3)), quantum_numbers=QN(0, 1), lineshape=constant_lineshape, argnames=[], preserve_partity=True, name='D_s0^*(2317)-')
        ],

        (1, 3): [
        Resonance(Node((1, 3)), quantum_numbers=QN(3, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=True, name='Xi_c(2815)0'),
            Resonance(Node((1, 3)), quantum_numbers=QN(1, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=True, name='Xi_c(2923)')
        ]
    }


    chains = [
        MultiChain(
            topology=topology,
            resonances = resonances,
            momenta = momenta,
            final_state_qn = final_state_qn
        ) for topology in topologies
    ]
    full = ChainCombiner(chains)
    # Start parameters for lineshape fitting
    start_parameters = {}
    
    # Add start parameters to the full object for easy access
    full.start_parameters = start_parameters

    return full

def test_Lb2LcD0K():



    import json
    import numpy as onp
    M_lb = 5.61951  # Mass of Lambda_b in GeV/c^2
    m_Lc = 2.28646  # Mass of Lambda_c+ in GeV/c^2
    m_D0 = 1.86484  # Mass of D0bar in GeV/c^2
    m_K = 0.49368   # Mass of K- in GeV/c^2

    with open("tests/test_data/Lb2LcD0K_momenta.json", "r") as f:
        momenta = json.load(f)
        momenta = {
            int(k): onp.array(v) for k, v in momenta.items()
        }
    full = amplitude(momenta)

    unpolarized, param_names = full.unpolarized_amplitude(full.generate_couplings())

    start_params = {
        name: full.start_parameters.get(name, 1.0) for name in param_names
    }

    print("Compiling unpolarized amplitude...")
    unpolarized = jit(unpolarized)
    unolarized_values = unpolarized(**start_params)
    print("Unpolarized amplitude compiled.")
    unolarized_values = onp.array(unolarized_values)
    
    # Calculate the weights
    weights_calc = onp.abs(onp.array(unolarized_values))**2 
    

if __name__ == "__main__":
    test_Lb2LcD0K()


from decayangle.decay_topology import Node
from decayangle.kinematics import mass
import numpy as np

def mass_from_node(node: Node, momenta):
    """
    Calculate the mass of a particle given its node and momenta.

    Parameters:
    - Node: The node representing the particle.
    - momenta: A dictionary containing the momenta of the particles.

    Returns:
    - The mass of the particle.
    """
    # copy the node
    node = Node(node.tuple)
    
    flat = tuple(np.array(node.tuple).flatten())

    return mass(
        sum(
            momenta[i] for i in flat
        )
    )



from typing import Union, Generator
from sympy import Rational, Symbol, lambdify
from sympy.physics.quantum.cg import CG
from sympy.physics.quantum.spin import Rotation
from decayamplitude.backend import numpy as np
from functools import lru_cache as cache
from sympy.abc import x as placeholder
from itertools import product

def convert_angular(f):
    """
    Wrapper to convert all passed values of type Angular to type int
    """
    def wrapped(*args, **kwargs):
        args = [arg.value2 if isinstance(arg, Angular) else arg for arg in args]
        kwargs = {key: value.value2 if isinstance(value, Angular) else value for key, value in kwargs.items()}
        return f(*args, **kwargs)
    return wrapped

class Angular:

    @classmethod
    def generate_helicities(cls, *angular_momenta:list["Angular"]) -> list[tuple[int]]:
        """
        Generate all possible helicities for a given set of angular momenta
        """
        return list(product(*[angular.projections(return_int=True) for angular in angular_momenta]))


    def __init__(self, angular_momentum:int):
        if not isinstance(angular_momentum, int):
            raise TypeError("Angular momentum must be an integer")

        self.angular_momentum = angular_momentum
    
    def __str__(self):
        return f"J={self.value}"
    
    def __hash__(self) -> int:
        return hash(self.angular_momentum)
    
    def __repr__(self):
        return self.__str__()
    
    def __eq__(self, other):
        return self.angular_momentum == other.angular_momentum

    @property
    def value(self):
        return self.angular_momentum / 2

    @property
    def value2(self):
        return self.angular_momentum
    
    @property
    def parity(self):
        if self.angular_momentum % 2 != 0:
            raise ValueError(f"Angular momentum must be even. {self} has no defined parity!")
        return (-1) ** int(self.value)

    def index(self):
        return self.angular_momentum

    def projections(self, return_int=False):
        """
        Returns the possible projections of the angular momentum
        """
        if return_int:
            return [i for i in range(-self.index(), self.index() + 1, 2)]
        return [Angular(i) for i in range(-self.index(), self.index() + 1, 2)]    
    
    def __add__(self, other):
        if isinstance(other, int):
            return Angular(self.angular_momentum + other)
        return Angular(self.angular_momentum + other.angular_momentum)
    
    def __sub__(self, other):
        return Angular(self.angular_momentum - other.angular_momentum)
    
    def couple(self, other):
        """
        Couple two angular momenta
        """
        minimum = abs(self.angular_momentum - other.angular_momentum)
        maximum = self.angular_momentum + other.angular_momentum
        return [Angular(i) for i in range(minimum, maximum + 1, 2)]



class QN:
    def __init__(self, angular_momentum:Union[int, Angular], parity: int) -> None:
        if isinstance(angular_momentum, int):
            self.angular = Angular(angular_momentum)
        else:
            self.angular = angular_momentum
        if not isinstance(parity, int):
            raise TypeError("Parity must be an integer not {}".format(type(parity)))
        if not parity in [-1, 1]:
            raise ValueError("Parity must be either -1 or 1 not {}".format(parity))
        self.parity = parity     

    def __str__(self):
        return f"{self.angular}^{self.parity}"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.angular == other.angular and self.parity == other.parity
    
    def __neq__(self, other):
        return not self == other

    def __add__(self, other):
        return QN(self.angular + other.angular, self.parity * other.parity)
    
    def __sub__(self, other):
        return QN(self.angular - other.angular, self.parity * other.parity)
    
    def couple(self, other):
        """
        Couple two quantum numbers
        """
        return [QN(j, p) for j in self.angular.couple(other.angular) for p in [self.parity * other.parity]]
    
    @classmethod
    def generate_L_states(cls, state0: "QN", state1:"QN", state2:"QN")-> Generator[tuple["QN", "QN"], None, None]:
        """
        state0 -> state1 + state2
        S = coupled spins of state 1 and state 2
        L = angular momentum of the 1-2 system
        L can be covered by coupling J with state0, since the possible L values arise from coupling J with a series of L values and looking which one can in turn couple to state0 spin. This is in pricinple due to detailed balance.

        params:
        state0 : QN
            Quantum numbers of the initial state
        state1 : QN
            Quantum numbers of the first final state
        state2 : QN
            Quantum numbers of the second final state

        returns:
        Generator[tuple[QN, QN]]
            Generator of possible (L, S) states
            Partity is not given, since it is defined by state0

        """
        for S in state1.angular.couple(state2.angular):
            for L in S.couple(state0.angular):
                if (L.parity * state1.parity * state2.parity) == state0.parity:
                    yield (L, S)
    
    def projections(self, return_int=False):
        return self.angular.projections(return_int=return_int)
        

@cache
def clebsch_gordan(j1, m1, j2, m2, J, M):
    """
    Return clebsch-Gordan coefficient. Note that all arguments should be multiplied by 2
    (e.g. 1 for spin 1/2, 2 for spin 1 etc.). Needs sympy.
    """

    cg = (
        CG(
            Rational(j1, 2),
            Rational(m1, 2),
            Rational(j2, 2),
            Rational(m2, 2),
            Rational(J, 2),
            Rational(M, 2),
        )
        .doit()
        .evalf()
    )
    cg = float(cg)
    if str(cg) == "nan":
        raise ValueError(f"CG({j1/2},{m1/2},{j2/2},{m2/2},{J/2},{M/2}) is not a number")
    return cg


@cache
def get_wigner_function(j: int, m1: int, m2: int):
    """
    Return Wigner small-d function. Note that all arguments should be multiplied by 2
    (e.g. 1 for spin 1/2, 2 for spin 1 etc.). Needs sympy.
    """
    j, m1, m2 = int(j), int(m1), int(m2)
    d = Rotation.d(Rational(j, 2), Rational(m1, 2), Rational(m2, 2), placeholder).doit().evalf()
    d = lambdify(placeholder, d, "numpy")
    return d

def wigner_small_d(theta, j, m1, m2):
    """Calculate Wigner small-d function. Needs sympy.
      theta : angle
      j : spin (in units of 1/2, e.g. 1 for spin=1/2)
      m1 and m2 : spin projections (in units of 1/2)

    :param theta:
    :param j:
    :param m1: before rotation
    :param m2: after rotation

    """
    d_func = get_wigner_function(j, m1, m2)
    d = d_func(theta)
    d = np.array(d, dtype=np.float64)
    # d[np.isnan(d)] = 0
    d = np.nan_to_num(d, copy=True, nan=0.0)
    d = d.astype(np.complex128)
    return d

@convert_angular
def wigner_capital_d(phi, theta, psi, j, m1, m2):
    return (
        np.exp(-1j * phi * m1 / 2)
        * wigner_small_d(theta, j, m1, m2)
        * np.exp(-1j * psi * m2 / 2)
    )


from utils import make_four_vectors, constant_lineshape, BW_lineshape
from decayamplitude.resonance import Resonance
from decayamplitude.rotation import QN, Angular
from decayamplitude.chain import DecayChain, MultiChain
from decayamplitude.combiner import ChainCombiner

from decayamplitude.backend import numpy as np

from decayangle.decay_topology import Topology, Node
from decayangle.kinematics import mass
from decayangle.config import config as decayangle_config
decayangle_config.backend = "jax"
decayangle_config.sorting = "off"

from jax import jit, grad

from collections import defaultdict

# Since we are semi leptonic there is only one important decay topology
# The (1, 2) system is the hadronic system, which is the one we are interested in
# Decay definition: B0 -> D0 h mu nu  or in the notation of the library 0 -> ((1, 2), (3, 4))
topology1 = Topology(
    0,
    decay_topology=((1,2), (3, 4))
)

def resonances_BW(momenta):
    m_12 = topology1.nodes[(1,2)].mass(momenta=momenta)
    resonances_hadronic = {
        (1,2): [
            # Here the hadronic resonances go+
            # These will decay strong, so we need to conserve parity
            Resonance(Node((1, 2)), quantum_numbers=QN(0, 1), lineshape=BW_lineshape(m_12), argnames=["D_2300_M", "D_2300_Gamma"], preserve_partity=True, name="D*0(2300)"),
            Resonance(Node((1, 2)), quantum_numbers=QN(4, 1), lineshape=BW_lineshape(m_12), argnames=["D_2460_M", "D_2460_Gamma"], preserve_partity=True, name="D*2(2460)"),
            Resonance(Node((1, 2)), quantum_numbers=QN(2, -1), lineshape=BW_lineshape(m_12), argnames=["D_2600_M", "D_2600_Gamma"], preserve_partity=True, name="D*1(2600)"),
            Resonance(Node((1, 2)), quantum_numbers=QN(0, 1), lineshape=constant_lineshape, argnames=[], preserve_partity=True, name="Non-Resonant-S-Wave"),
            # Resonance(Node((1, 2)), quantum_numbers=QN(J, P), lineshape=BW_lineshape(m_12), argnames=["mass_resonance_n", "width_resonance_n"], preserve_partity=True), # template for further resonances
            ],
        # This is the W boson. It is defined as a resonance, but we assue a constant lineshape in this mass regime. One could use a more complicated one aswell.
        (3, 4): [Resonance(Node((3, 4)), quantum_numbers=QN(2, -1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="W")],

        # This is the decaying B0 meson. It is defined as a resonance, but since this is a decay amplitude, the description is not important. Only the QN have to be correct. 
        0: [Resonance(Node(0), quantum_numbers=QN(0, 1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="B0")],
    }
    return resonances_hadronic


def phasespace_momenta():
    # dependence on the phasespace library, dont use, if you dont have it installed
    # Installing via pip may destroy you existing  cuda setup
    # This is a simple example, where we use the phasespace library to generate the momenta
    import phasespace
    B0_MASS = 5.27963e3
    PION_MASS = 0.13957018e3
    D0_MASS = 1.86483e3
    MU_MASS = 0.1056583715e3
    NU_MASS = 0
    weights, particles = phasespace.nbody_decay(B0_MASS,
                                            [D0_MASS, PION_MASS, MU_MASS, NU_MASS]).generate(n_events=100000)
    momenta = {
        1: np.array(particles["p_0"]),
        2: np.array(particles["p_1"]),
        3: np.array(particles["p_2"]),
        4: np.array(particles["p_3"]),
    }
    return momenta

def shortFourBodyAmplitudeBW():
    final_state_qn = {
        1: QN(0, 1), # D0
        2: QN(0, 1), # h
        3: QN(1, 1), # mu
        4: QN(1, -1) # nu
    }
    momenta = phasespace_momenta()

    resonances_hadronic = resonances_BW(momenta)
    chain1 = MultiChain(
        topology = topology1,
        resonances = resonances_hadronic,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    # For semileptonics no actual alignment is needed, as we are only interested in one single decay chain
    # But doing it this way allows to use the convenience functions of the combiner. For a one chais setup the combiner will not align anyways.
    full = ChainCombiner([chain1])

    # The unpolarized amplitude is the simplest one, and the default case in LHCb
    unpolarized, argnames = full.unpolarized_amplitude(
        full.generate_couplings() # This is a helper function to generate the couplings for the hadronic system, if you want to restrict them, you will have to do it manually.
                                    # Alternatively you can also restrict the couplings in the fitter later.      
        )
    # argnames are the names of the arguments of the function, which are the masses and widths of the resonances and the couplings
    # The order of argnames is the order of the arguments in the function
    # The function can be called with positional arguments, or with keyword arguments
    # so unpolarized(*[1, 2, 3, 4, 5, 6]) is the same as unpolarized(mass_resonance_1=1, width_resonance_1=2, mass_resonance_2=3, width_resonance_2=4, mass_resonance_3=5, width_resonance_3=6), omitting the couplings
    print(argnames)

    # an issue with jax, where the internal caching structure needs to be prewarmed, so that in the compilation step the correct types are inferred
    print(unpolarized(*([1.0] * len(argnames))))
    
    # we can now jit the function, to make it faster after the compile
    unpolarized = jit(unpolarized) 
    print(unpolarized(*([1.0] * len(argnames))))


    # for the gradient calculation we need to define a log likelihood function or something, that produces a single value
    def LL(*args):
        return np.sum(
            np.log(unpolarized(*args))
                )
    # we can calc the gradient of the log likelihood function
    unpolarized_grad = jit(grad(LL, argnums=[i for i in range(len(argnames))]))

    # and a test call (may take quite some time)
    # print(unpolarized_grad(*([1.0] * len(argnames))))

    # Further calls will be much faster, since we dont need to compile again
    # print(unpolarized_grad(*([1.0] * len(argnames))))
    # print(unpolarized_grad(*([2.0] * len(argnames))))


    # Other options for amplitudes, one might be interested in
    # polarized, lambdas ,polarized_argnames = full.polarized_amplitude(full.generate_couplings())
    # print(lambdas)
    # lambda_values = [0, 0, 0, 1, 1]
    # print(polarized(*lambda_values,*([1] * len(polarized_argnames))) )

    # matrix_function, matrix_argnames = full.matrix_function(full.generate_couplings())
    # print(matrix_argnames)
    # print(matrix_function(0, *([1] * len(argnames))) )


    # # quick plot to show the resonant structure can be confiremd visually
    # import matplotlib.pyplot as plt
    # # set couplings to 1.0
    # param_dict = {param_name: 1.0 for param_name in argnames}
    # param_dict.update({
    #     # made up masses and widths
    #     'D_2300_Gamma': 20, 
    #     'D_2460_Gamma': 60, 
    #     'D_2460_M': 2460, 
    #     'D_2300_M': 2300, 
    #     'D_2600_M': 2600, 
    #     'D_2600_Gamma': 80,
    #     # set the non resonant part to a low value, since otherwise it will dominate the plot
    #     "Non_Resonant_S_Wave_to_particle_1_particle_2_LS_0_0_real": 0.00001,
    #     "Non_Resonant_S_Wave_to_particle_1_particle_2_LS_0_0_imaginary": 0.0,
    # })
    # del param_dict['Non_Resonant_S_Wave_to_particle_1_particle_2_LS_0_0_real']
    # import numpy as onp # original numpy plays nicer with matplotlib, so we convert the results to numpy
    # weights = onp.array(unpolarized(**param_dict))
    # hadronic_mass = topology1.nodes[(1,2)].mass(momenta=momenta)
    # plt.hist(onp.array(hadronic_mass), bins=100, weights=weights, label="Hadronic mass")
    # plt.show()



if __name__ == "__main__":
    shortFourBodyAmplitudeBW()


from decayangle.decay_topology import Topology, HelicityAngles, WignerAngles
from decayangle.lorentz import LorentzTrafo
from decayamplitude.backend import numpy as np

def constant_lineshape(*args):
    return 1

def BW_lineshape(mass):
    # usually the function is called for a fixed set of points in the phasespace. 
    # Thus we can give the mass values as fixed
    # decayamplitude will provide l and s values per default for the resonances, but not the mass values
    def BW(l,s, m0, gamma):
        return 1 / (mass**2 - m0**2 + 1j * mass * gamma) 
    return BW

def make_four_vectors(phi_rf, theta_rf, psi_rf):
    from decayamplitude.backend import numpy as np

    # Make sure, the sorting is turned off

    # Given values
    # Lc -> p K pi
    m0 = 6.32397
    m12 = 9.55283383**0.5
    m23 = 26.57159046**0.5
    m13 = 17.86811729**0.5
    m1, m2, m3 = 1, 2, 3
    # Squared masses
    m0sq, m1sq, m2sq, m3sq, m12sq, m23sq = [x**2 for x in [m0, m1, m2, m3, m12, m23]]

    # Källén function
    def Kallen(x, y, z):
        return x**2 + y**2 + z**2 - 2 * (x * y + x * z + y * z)

    # Calculating missing mass squared using momentum conservation
    m31sq = m0sq + m1sq + m2sq + m3sq - m12sq - m23sq

    # Momenta magnitudes
    p1a = np.sqrt(Kallen(m23sq, m1sq, m0sq)) / (2 * m0)
    p2a = np.sqrt(Kallen(m31sq, m2sq, m0sq)) / (2 * m0)
    p3a = np.sqrt(Kallen(m12sq, m3sq, m0sq)) / (2 * m0)

    # Directions and components
    cos_zeta_12_for0_numerator = (m0sq + m1sq - m23sq) * (
        m0sq + m2sq - m31sq
    ) - 2 * m0sq * (m12sq - m1sq - m2sq)
    cos_zeta_12_for0_denominator = np.sqrt(Kallen(m0sq, m2sq, m31sq)) * np.sqrt(
        Kallen(m0sq, m23sq, m1sq)
    )
    cos_zeta_12_for0 = cos_zeta_12_for0_numerator / cos_zeta_12_for0_denominator

    p1z = -p1a
    p2z = -p2a * cos_zeta_12_for0
    p2x = np.sqrt(p2a**2 - p2z**2)
    p3z = -p2z - p1z
    p3x = -p2x

    # Energy calculations based on the relativistic energy-momentum relation
    E1 = np.sqrt(p1z**2 + m1sq)
    E2 = np.sqrt(p2z**2 + p2x**2 + m2sq)
    E3 = np.sqrt(p3z**2 + p3x**2 + m3sq)

    # Vectors
    p1 = np.array([0, 0, p1z, E1])
    p2 = np.array([p2x, 0, p2z, E2])
    p3 = np.array([p3x, 0, p3z, E3])

    # Lorentz transformation
    momenta = {i: p for i, p in zip([1, 2, 3], [p1, p2, p3])}
    tree1 = Topology(root=0, decay_topology=((2, 3), 1))

    # momenta = Topology(root=0, decay_topology=((1, 2), 3)).align_with_daughter(momenta, 3)
    # momenta = tree1.root.transform(LorentzTrafo(0, 0, 0, 0, -np.pi, 0), momenta)
    rotation = LorentzTrafo(0, 0, 0, phi_rf, theta_rf, psi_rf)

    momenta_23_rotated = tree1.root.transform(rotation, momenta)
    return momenta_23_rotated


from decayangle.decay_topology import Topology, TopologyCollection, HelicityAngles
from decayangle.lorentz import LorentzTrafo
from decayangle.config import config as decayangle_config
import numpy as np

def make_numpy(f):
    def wrapper(*args, **kwargs):
        args = [np.array(arg) for arg in args]
        kwargs = {k: np.array(v,dtype=np.float64) for k, v in kwargs.items()}
        return f(*args, **kwargs)
    return wrapper

@make_numpy
def make_four_vectors_from_dict(mkpisq, mkpsq, mppisq, phip, thetap, chi, phi_Kst = None, theta_Kst=None, phi_K=None, theta_K=None):
    import numpy as np


    # Make sure, the sorting is turned off

    # Given values
    # Lc -> p K pi
    # 0 -> 1 2 3
    m12 = np.sqrt(mkpsq)
    m23 = np.sqrt(mkpisq)
    m13 = np.sqrt(mppisq)
    m1, m2, m3 = 0.938272088, 0.493677, 0.13957039
    m0 = np.sqrt((mkpisq + mkpsq + mppisq) - m1**2  - m2**2 - m3**2)

    # Squared masses
    m0sq, m1sq, m2sq, m3sq, m12sq, m23sq = [x**2 for x in [m0, m1, m2, m3, m12, m23]]

    # Källén function
    def Kallen(x, y, z):
        return x**2 + y**2 + z**2 - 2 * (x * y + x * z + y * z)

    # Calculating missing mass squared using momentum conservation
    m31sq = m0sq + m1sq + m2sq + m3sq - m12sq - m23sq

    # Momenta magnitudes
    p1a = np.sqrt(Kallen(m23sq, m1sq, m0sq)) / (2 * m0)
    p2a = np.sqrt(Kallen(m31sq, m2sq, m0sq)) / (2 * m0)

    # Directions and components
    cos_zeta_12_for0_numerator = (m0sq + m1sq - m23sq) * (
        m0sq + m2sq - m31sq
    ) - 2 * m0sq * (m12sq - m1sq - m2sq)
    cos_zeta_12_for0_denominator = np.sqrt(Kallen(m0sq, m2sq, m31sq)) * np.sqrt(
        Kallen(m0sq, m23sq, m1sq)
    )
    cos_zeta_12_for0 = cos_zeta_12_for0_numerator / cos_zeta_12_for0_denominator

    p1z = -p1a
    p2z = -p2a * cos_zeta_12_for0
    p2x = np.sqrt(p2a**2 - p2z**2)
    p3z = -p2z - p1z
    p3x = -p2x

    # Energy calculations based on the relativistic energy-momentum relation
    E1 = np.sqrt(p1z**2 + m1sq)
    E2 = np.sqrt(p2z**2 + p2x**2 + m2sq)
    E3 = np.sqrt(p3z**2 + p3x**2 + m3sq)

    # Vectors such that we align with proton momentum
    p1 = np.array([0, 0, p1z, E1])
    p2 = np.array([p2x, 0, p2z, E2])
    p3 = np.array([p3x, 0, p3z, E3])

    # Lorentz transformation
    momenta = {i: p for i, p in zip([1, 2, 3], [p1, p2, p3])}
    tree1 = Topology(root=0, decay_topology=((2, 3), 1))
    # momenta are now in x-z plane

    phip = -np.pi + phip 

    thetap = np.pi - thetap
    chi = -np.pi + chi
    # rotation = LorentzTrafo(0, 0, 0, phip, thetap, chi)
    rotation = LorentzTrafo(0, 0, 0, phi_Kst, theta_Kst, phi_K)

    momenta_23_rotated = tree1.root.transform(rotation, momenta)
    return momenta_23_rotated


decayangle_config.sorting = "off"


from decayamplitude.chain import DecayChain, AlignedChain, MultiChain, AlignedMultiChain
from decayangle.decay_topology import Topology
from typing import Union, Callable
from decayamplitude.resonance import LSTuple, Resonance
from decayamplitude.utils import _create_function

class ChainCombiner:
    """
    Class to automatically combine multiple decay chains into a single amplitude.
    The first chain is used as a reference for the topology.
    All other chains will be transformed into the reference basis.
    """

    def __init__(self, chains: list[Union[DecayChain, MultiChain]]) -> None:
        self.chains = chains
        self.reference = chains[0]
        self.aligned_chains = [
            AlignedMultiChain.from_multichain(
                chain,
                self.reference
            ) if isinstance(chain, MultiChain) else 
                AlignedChain(
                    chain.topology,
                    chain.resonances,
                    chain.momenta,
                    chain.final_state_qn,
                    self.reference
                )
            for chain in chains[1:]
        ]


    @property
    def root_resonance(self):
        if all(chain.root_resonance.quantum_numbers == self.reference.root_resonance.quantum_numbers for chain in self.chains):
            return self.reference.root_resonance
        return None


    @property
    def combined_function(self):
        """
        Returns a function that combines the amplitudes of all chains
        """
        def f(h0, lambdas:dict, arguments:dict):

            amplitudes = [
                chain.aligned_matrix(h0, arguments)[tuple(lambdas[k] for k in sorted(lambdas.keys()))]
                for chain in self.aligned_chains
            ]
            return sum(amplitudes) + self.reference.chain_function(h0, lambdas, arguments)

        return f
    
    def polarized_amplitude(self, ls_couplings:dict[int, dict[str: dict[LSTuple, float]]]) -> tuple[Callable, list[str], list[str]]:
        """
        Returns a function that combines the amplitudes of all chains
        """
        sorted_final_state_nodes = sorted([n.node.value for n in self.reference.final_state_nodes])
        final_state_lambdas = sorted([f"h_{n}" for n in sorted_final_state_nodes]) 
        def fun(arguments:dict):
            # build lambda dict, as it is used internally from plain parameters
            h0 = arguments.pop("h0")
            lambdas = {n: arguments.pop(k) for k, n  in zip(final_state_lambdas, sorted_final_state_nodes)}
            return self.combined_function(h0, lambdas, arguments)
        polarized, argnames = _create_function(["h0", *final_state_lambdas] + self.resonance_params, ls_couplings, fun)

        return polarized, ["h0", *final_state_lambdas], argnames[len(final_state_lambdas)+1:]

    
    @property
    def combined_matrix(self) -> Callable:
        """
        Returns a function that combines the matrices of all chains.
        The final matrix will be a sum of all matrices, where the alignment is already performed.
        """
        def matrix(h0, arguments:dict) -> dict:
            matrices = [
                chain.aligned_matrix(h0, arguments)
                for chain in self.aligned_chains
            ]
            matrices.append(self.reference.matrix(h0, arguments))

            return {
                key: sum(matrix[key] for matrix in matrices)
                for key in matrices[0].keys()
            }
        return matrix
    
    def matrix_function(self, ls_couplings:dict[int, dict[str: dict[LSTuple, float]]], complex_couplings: bool=True) -> tuple[Callable, list[str]]:
        """
        Returns a function that combines the matrices of all chains.
        The final matrix will be a sum of all matrices, where the alignment is already performed.
        """
        if "h0" in self.resonance_params:
            raise ValueError("The parameter name 'h0' is reserved for the helicity quantum number of the mother particle. Please choose another name for the resonance parameter.")
        def fun(arguments:dict):
            h0 = arguments["h0"]
            return self.combined_matrix(h0, arguments)
        return _create_function(["h0"] + self.resonance_params, ls_couplings, fun, complex_couplings=complex_couplings)
    
    def generate_couplings(self):
        """
        Generates the couplings for the ls basis.
        """
        couplings = {}
        for chain in self.chains:
            couplings.update(chain.generate_couplings())
        return couplings
    
    @property
    def resonance_params(self) -> list[str]:
        resonance_parameter_names = [name for chain in self.chains for name in chain.resonance_params]

        if len(set(resonance_parameter_names)) != len(resonance_parameter_names):
            from collections import Counter
            c = Counter(resonance_parameter_names)
            # raise ValueError(f"Parameter names are not unique: {', '.join([name for name, count in c.items() if count > 1])}")
        return list(set(resonance_parameter_names))

    def unpolarized_amplitude(self, ls_couplings: dict, complex_couplings=True) -> tuple[Callable, list[str]]:

        if self.root_resonance is None:
            raise ValueError(f"The root resonance must be the same for all chains! Root = {self.reference.topology.root}.")

        def f(arguments:dict):
            return sum(
                    abs(v)**2 
                    for h0 in self.root_resonance.quantum_numbers.angular.projections()
                    for v in self.combined_matrix(h0, arguments).values()
                )

        return _create_function(self.resonance_params, ls_couplings, f, complex_couplings=complex_couplings)
        

        




from generate_momenta import make_four_vectors_from_dict
from itertools import product
from decayangle.decay_topology import Topology, TopologyCollection, HelicityAngles
from decayangle.config import config as decayangle_config
import numpy as np 

from decayamplitude.resonance import Resonance
from decayamplitude.rotation import QN
from decayamplitude.chain import AlignedChain
from decayamplitude.combiner import ChainCombiner

from decayamplitude.backend import numpy as np



from decayangle.decay_topology import Topology, Node
from decayangle.kinematics import mass
from decayangle.config import config as decayangle_config


def constant_lineshape(*args):
    return 1

decayangle_config.sorting = "off"

tg = TopologyCollection(
    0,
    topologies=[
        Topology(0, decay_topology=((2, 3), 1)),
        Topology(0, decay_topology=((3, 1), 2)),
        Topology(0, decay_topology=((1, 2), 3)),
    ],
)


# Lc -> p K pi
# 0 -> 1 2 3

def read_helicity_angles_from_dict(dtc):
    mappings = {
        ((2, 3), 1): ("Kpi", "theta_Kst", "phi_Kst", "theta_K", "phi_K"),
        ((3, 1), 2): ("pip", "theta_D", "phi_D", "theta_pi", "phi_pi"),
        ((1, 2), 3): ("pK", "theta_L", "phi_L", "theta_p", "phi_p"),
    }

    topos = {}

    for tpl, (name, theta_hat, phi_hat, theta, phi) in mappings.items():
        topos[tpl] = {
            tpl: HelicityAngles(
                dtc[name][phi_hat],
                dtc[name][theta_hat],
            ),
            tpl[0]: HelicityAngles(
                dtc[name][phi],
                dtc[name][theta],
            ),
        }
    return topos


class Amplitude:
    def __init__(self, momenta):
        self.momenta = momenta
        L_1520 = Resonance(Node((1, 2)), quantum_numbers=QN(3, -1), lineshape=constant_lineshape, argnames=[], name = "L_1520", preserve_partity=True, scheme="helicity") # , scheme="helicity"
        Lc = Resonance(Node(0), quantum_numbers=QN(1, 1), lineshape=constant_lineshape, argnames=[], preserve_partity=False, name="Lc", scheme="helicity") # , scheme="helicity"
        self.resonances =   {
            (1, 2): L_1520,
            0: Lc,
        }
        self.final_state_qn = {
            1: QN(1, 1),
            2: QN(0, 1),
            3: QN(0, 1),
        }
        self.topology = Topology(0, decay_topology=((1, 2), 3))
        self.reference_topology = Topology(0, decay_topology=((2, 3), 1))
        self.helicity_angles = self.topology.helicity_angles(momenta=self.momenta)
        self.chain = AlignedChain(
                topology = self.topology,
                reference= self.reference_topology,
                resonances = self.resonances,
                momenta = momenta,
                final_state_qn = self.final_state_qn,
                convention="helicity",
            )
        
        self.combiner = ChainCombiner([self.chain, ])

        matrix_function, matrix_argnames = self.chain.aligned_matrix_function(self.chain.generate_couplings())
        
        couplings_m1_m1 = {
            "Lc_H_-3_0": 0,
            "Lc_H_-1_0": 1,
            "Lc_H_1_0": 0,
            "Lc_H_3_0": 0,
            "L_1520_H_1_0": 0,
            "L_1520_H_-1_0":  1/ (4)**0.5,
        }
        couplings_1_m1 = {
            "Lc_H_-3_0": 0,
            "Lc_H_-1_0": 0,
            "Lc_H_1_0": 1,
            "Lc_H_3_0": 0,
            "L_1520_H_1_0": 0,
            "L_1520_H_-1_0": 1/ (4)**0.5,
        }
        couplings_m1_1 = {
            "Lc_H_-3_0": 0,
            "Lc_H_-1_0": 1,
            "Lc_H_1_0": 0,
            "Lc_H_3_0": 0,
            "L_1520_H_1_0":  1/ (4)**0.5,
            "L_1520_H_-1_0": 0,
        }
        couplings_1_1 = {
            "Lc_H_-3_0": 0,
            "Lc_H_-1_0": 0,
            "Lc_H_1_0": 1,
            "Lc_H_3_0": 0,
            "L_1520_H_1_0":  1/ (4)**0.5,
            "L_1520_H_-1_0": 0,
        }

        arguments_1_1 = {
            L_1520.id: {
                    "couplings": {
                (1, 0) : 1/ (4)**0.5,
                (-1, 0) : 0,
            }},
            Lc.id: {
                    "couplings": {
                (1, 0) : 1,
                (-1, 0) : 0,
                (3, 0) : 0,
                (-3, 0) : 0,
            }
            }
        }
        arguments_m1_1 = {
            L_1520.id: {
                    "couplings": {
                (1, 0) : 1/ (4)**0.5,
                (-1, 0) : 0,
            }},
            Lc.id: {
                    "couplings": {
                (1, 0) : 0,
                (-1, 0) : 1,
                (3, 0) : 0,
                (-3, 0) : 0,
            }
            }
        }

        arguments_1_m1 = {
            L_1520.id: {
                    "couplings": {
                (1, 0) : 0,
                (-1, 0) : 1/ (4)**0.5,
            }
            },
            Lc.id: {
                    "couplings": {
                (1, 0) : 1,
                (-1, 0) : 0,
                (3, 0) : 0,
                (-3, 0) : 0,
            }
            }
        }

        arguments_m1_m1 = {
            L_1520.id: {
                    "couplings": {
                    (1, 0) : 0,
                    (-1, 0) : 1 / (4)**0.5,
                }
            },
            Lc.id:{
                "couplings":  {
                (1, 0) : 0,
                (-1, 0) : 1,
                (3, 0) : 0,
                (-3, 0) : 0,
            }
            }
        }




        aligned_matrix = self.chain.aligned_matrix
        # (hlc, l1520)
        self.value = {
            ( 1, -1, -1): matrix_function( 1, **couplings_m1_m1),
            ( 1, -1,  1): matrix_function( 1, **couplings_m1_1),
            ( 1,  1, -1): matrix_function( 1, **couplings_1_m1),
            ( 1,  1,  1): matrix_function( 1, **couplings_1_1),

            (-1, -1, -1): matrix_function(-1, **couplings_m1_m1),
            (-1, -1,  1): matrix_function(-1, **couplings_m1_1),
            (-1,  1, -1): matrix_function(-1, **couplings_1_m1),
            (-1,  1,  1): matrix_function(-1, **couplings_1_1)
        }
        # self.value = {
        #     (1, -1, -1): aligned_matrix(1, arguments_m1_m1),
        #     (1, -1, 1): aligned_matrix(1, arguments_m1_1),
        #     (1, 1, -1): aligned_matrix(1, arguments_1_m1),
        #     (1, 1, 1): aligned_matrix(1, arguments_1_1),

        #     (-1, -1, -1): aligned_matrix(-1, arguments_m1_m1),
        #     (-1, -1, 1): aligned_matrix(-1, arguments_m1_1),
        #     (-1, 1, -1): aligned_matrix(-1, arguments_1_m1),
        #     (-1, 1, 1): aligned_matrix(-1, arguments_1_1)
        # }

    @property
    def wig_d(self):
        Lc_decay_angle = self.helicity_angles[((1, 2), 3)]
        L1520_decay_angle = self.helicity_angles[(1,2)]
        Lc_J = 1
        L1520_J = 3

        return {
            (h_lc, h_l1520, h_p): wigner_capital_d(Lc_decay_angle.theta_rf, Lc_decay_angle.phi_rf, 0, Lc_J, h_lc, h_l1520) * wigner_capital_d(L1520_decay_angle.theta_rf, L1520_decay_angle.phi_rf, 0, L1520_J, h_l1520, h_p)
            for h_lc, h_l1520, h_p in product([1, -1], [1, -1], [1, -1])
        }


def parse_complex(s):
    try: 
        return complex(s)
    except Exception as e:
        print(s)
        raise e


def get_result_amplitude(dtc, a, b, c, d):
    key = f"L(1520)_{{{a}, {b}}}"
    dtc = dtc[key]

    dtc = {
        k: parse_complex(v.replace("im", "j").replace("+ -", "-").replace(" ", ""))
        for k, v in dtc.items()
    }
    return dtc[f"A[{c},{d}]"], f"L(1520)_{{{a}, {b}}} A[{c},{d}]"


def r_phi(comp):
    return f"{abs(comp)} {float(np.angle(comp))}"

from decayamplitude.rotation import wigner_capital_d
def test_elisabeth():
    import json

    path = "examples/test_data/Parsed_ccp_kinematics_100events.json"
    result_path = "examples/test_data/cpp_100_events_sign2pi_unmodified.json"
    with open(path, "r") as f:
        data = json.load(f)
    with open(result_path, "r") as f:
        result = json.load(f)

    
    for k, dtc in list(data.items()):
        kwargs = {k: v for k, v in dtc["kinematic"].items() if k != "mkpisq" }
        momenta = make_four_vectors_from_dict(**dtc["chain_variables"]["Kpi"], **kwargs)
        amplitude = Amplitude(momenta)
        for hl1520, pi, hlc, hp in product([1, -1], [0], [1, -1], [1, -1]):
            el, string = get_result_amplitude(result[k],  hl1520, pi, hlc, hp)
            kai = amplitude.value[(hlc, hl1520, hp)][(hp, 0, 0)]
            kai = amplitude.wig_d[(hlc, hl1520, hp)]
            if not np.allclose(el, kai, atol=0.04):
                print(f"{hlc=}, {hl1520=},  {hp=} string = {string}")
                print("Rphi",r_phi(el), r_phi(kai))
                print("value",el, kai)
                print("Ratio", el / kai)
                print()
            else:
                # continue
                print("OK", f"{hl1520=}, {hlc=}, {hp=} string = {string}")

           
            # print(get_result_amplitude(result[k], hlc , pi, hl1520, hp))


        
        # res1 = get_result_amplitude(result[k], 1, 0)
        # res2 = get_result_amplitude(result[k], -1, 0)

        exit(0)


if __name__ =="__main__":
    test_elisabeth()


from utils import make_four_vectors, constant_lineshape, BW_lineshape
from decayamplitude.resonance import Resonance
from decayamplitude.rotation import QN, Angular
from decayamplitude.chain import DecayChain, MultiChain
from decayamplitude.combiner import ChainCombiner

from decayamplitude.backend import numpy as np



from decayangle.decay_topology import Topology, Node
from decayangle.kinematics import mass
from decayangle.config import config as decayangle_config

from collections import defaultdict

# we want to define our chains by hand, so we need to turn off the automatic sorting
decayangle_config.sorting = "off" 

def resonances() -> tuple[dict]:
    resonances1 = {
        (2,3): Resonance(Node((2, 3)), quantum_numbers=QN(0, -1), lineshape=constant_lineshape, argnames=[]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[], preserve_partity=False)
    }

    resonances2 = {
        (1, 2): Resonance(Node((1, 2)), quantum_numbers=QN(3, -1), lineshape=constant_lineshape, argnames=[]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[], preserve_partity=False)
    }

    resonances3 = {
        (1, 2): Resonance(Node((1, 2)), quantum_numbers=QN(1, -1), lineshape=constant_lineshape, argnames=[]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[], preserve_partity=False)
    }

    resonances_dpd = {
        (2,3): Resonance(Node((2, 3)), quantum_numbers=QN(4, -1), lineshape=constant_lineshape, argnames=[]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[], preserve_partity=False)
    }

    return resonances1, resonances2, resonances3, resonances_dpd

def resonances_BW(momenta) -> tuple[dict]:
    nodes = Topology(0, decay_topology=((2,3), 1)).nodes
    nodes_2 = Topology(0, decay_topology=((1,2), 3)).nodes

    resonances1 = {
        (2,3): Resonance(nodes[(2, 3)], quantum_numbers=QN(0, -1), lineshape=BW_lineshape(nodes[(2, 3)].mass(momenta)), argnames=["gamma1", "m01"]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[], preserve_partity=False)
    }

    resonances2 = {
        (1, 2): Resonance(nodes_2[(1, 2)], quantum_numbers=QN(3, -1), lineshape=BW_lineshape(nodes_2[(1, 2)].mass(momenta)), argnames=["gamma2", "m01"]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[], preserve_partity=False)
    }

    resonances3 = {
        (1, 2): Resonance(Node((1, 2)), quantum_numbers=QN(1, -1), lineshape=BW_lineshape(nodes_2[(1, 2)].mass(momenta)), argnames=["gamma1", "m01"]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[], preserve_partity=False)
    }

    resonances_dpd = {
        (2,3): Resonance(nodes[(2, 3)], quantum_numbers=QN(4, -1), lineshape=BW_lineshape(nodes[(2, 3)].mass(momenta)), argnames=["gamma1", "m01"]),
        0: Resonance(Node(0), 1, 1, lineshape=constant_lineshape, argnames=[], preserve_partity=False)
    }

    return resonances1, resonances2, resonances3, resonances_dpd

def threeBodyAmplitude():
    """
    Whe create an amplitdue for a three-body decay chain with the following topology:
    0 -> ((2,3)-> 2 3) 1
    
    """
    topology1 = Topology(
        0,
        decay_topology=((2,3), 1)
    )

    topology2 = Topology(
        0,
        decay_topology=((1, 2), 3)
    )

    # we need to define the momenta for the decay chain
    momenta = make_four_vectors(1,2,np.linspace(0,np.pi,10))

    final_state_qn = {
            1: QN(1, 1),
            2: QN(2, 1),
            3: QN(0, 1)
        }
    resonances1, resonances2, resonances3, resonances_dpd = resonances()
    decay = DecayChain(
        topology = topology1,
        resonances = resonances1,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    decay2 = DecayChain(
        topology = topology2,
        resonances = resonances2,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    decay3 = DecayChain(
        topology = topology2,
        resonances = resonances3,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    decay_dpd = DecayChain(
        topology = topology1,
        resonances = resonances_dpd,
        momenta = momenta,
        final_state_qn = final_state_qn
    )

    arguments1 = {
        resonances1[(2,3)].id : {
            "couplings":{
                (2, 2) : 1
            }
        }, 
        resonances1[0].id : {
            "couplings":{
                (2, 1) : 1
            }
        }
    }
    arguments2 = {
        resonances2[(1,2)].id : {
            "couplings":{
                (2, 3) : 1
            }
        }, 
        resonances2[0].id : {
            "couplings":{
                (2, 3) : 1
            }
        }
    }

    arguments3 = {
        resonances3[(1,2)].id : {
            "couplings":{
                (2, 3) : 1
            }
        }, 
        resonances3[0].id : {
            "couplings":{
                (2, 1) : 1
            }
        }
    }

    arguments_dpd = {
        resonances_dpd[(2,3)].id : {
            "couplings": {
                (4, 2): 1
            },
        },
        resonances_dpd[0].id : {
            "couplings": {
                (4, 3): 1
            }
        }
    }

    arguments = {}
    arguments.update(arguments1)
    arguments.update(arguments2)
    arguments.update(arguments3)
    arguments.update(arguments_dpd)


    full = ChainCombiner([decay, decay2, decay3, decay_dpd])
    print(full.combined_function(0, {1:1, 2:0, 3:0} ,arguments))

    all_helicities = Angular.generate_helicities(*[final_state_qn[key].angular for key in final_state_qn.keys()])
    all_helicities = [
        {key: helicity[i] for i, key in enumerate(final_state_qn.keys())}
        for helicity in all_helicities
    ]

    full_matrix_1 = full.combined_matrix(-1, arguments)
    full_matrix_2 = full.combined_matrix(1, arguments)
    print(sum(abs(v)**2 for v in full_matrix_1.values()) + 
          sum(abs(v)**2 for v in full_matrix_2.values()))
    

    # Advantage of this setup is, that we only have one alignment step per topology
    ch1 = MultiChain.from_chains([decay, decay_dpd])
    ch2 = MultiChain.from_chains([decay2, decay3])
    full_multi = ChainCombiner([ch1, ch2])
    full_matrix_multi_1 = full_multi.combined_matrix(-1, arguments)
    full_matrix_multi_2 = full_multi.combined_matrix(1, arguments)
    print(sum(abs(v)**2 for v in full_matrix_multi_1.values()) + 
          sum(abs(v)**2 for v in full_matrix_multi_2.values())
    )


def shortThreeBodyAmplitude():
    """
    We can also combine the chains in a shorter way, by using the MultiChain class
    """
    final_state_qn = {
            1: QN(1, 1),
            2: QN(2, 1),
            3: QN(0, 1)
        }
    resonances1, resonances2, resonances3, resonances_dpd = resonances()
    momenta = make_four_vectors(1,2,np.linspace(0,np.pi,10))
    topology1 = Topology(
        0,
        decay_topology=((2,3), 1)
    )
    topology2 = Topology(
        0,
        decay_topology=((1, 2), 3)
    )

    chain1 = MultiChain.from_chains([
        DecayChain(
            topology = topology1,
            resonances = resonances1,
            momenta = momenta,
            final_state_qn = final_state_qn
        ),
        DecayChain(
            topology = topology1,
            resonances = resonances_dpd,
            momenta = momenta,
            final_state_qn = final_state_qn
        )
    ])

    chain2 = MultiChain.from_chains([
        DecayChain(
            topology = topology2,
            resonances = resonances2,
            momenta = momenta,
            final_state_qn = final_state_qn,
        ),
        DecayChain(
            topology = topology2,
            resonances = resonances3,
            momenta = momenta,
            final_state_qn = final_state_qn
        )
    ])

    merged_resonances = defaultdict(list)
    for key, resonance in resonances3.items():
        merged_resonances[key].append(resonance)
    for key, resonance in resonances2.items():
        merged_resonances[key].append(resonance)
    merged_resonances = dict(merged_resonances)

    merged_resonances[0] = [merged_resonances[0][0]]

    full = ChainCombiner([chain1, chain2])
    arguments = full.generate_couplings()

    matrix1 = full.combined_matrix(-1, arguments)
    matrix2 = full.combined_matrix(1, arguments)
    print(sum(abs(v)**2 for v in matrix1.values()) + 
          sum(abs(v)**2 for v in matrix2.values())
    )
    unpolarized, argnames = full.unpolarized_amplitude(full.generate_couplings())
    print(argnames)
    print(unpolarized(*([1] * len(argnames))))
    full2 = ChainCombiner([chain1, MultiChain(topology2, momenta=momenta, resonances=merged_resonances, final_state_qn=final_state_qn)])
    unpolarized, argnames = full2.unpolarized_amplitude(full2.generate_couplings())
    print(argnames)
    print(unpolarized(*([1] * len(argnames))))


def shortThreeBodyAmplitudeBW():
    final_state_qn = {
        1: QN(1, 1),
        2: QN(2, 1),
        3: QN(0, 1)
    }
    momenta = make_four_vectors(1,2,np.linspace(0,np.pi,10))
    topology1 = Topology(
        0,
        decay_topology=((2,3), 1)
    )
    topology2 = Topology(
        0,
        decay_topology=((1, 2), 3)
    )
    resonances1, resonances2, resonances3, resonances_dpd = resonances_BW(momenta)
    chain1 = MultiChain.from_chains([
        DecayChain(
            topology = topology1,
            resonances = resonances1,
            momenta = momenta,
            final_state_qn = final_state_qn
        ),
        DecayChain(
            topology = topology1,
            resonances = resonances_dpd,
            momenta = momenta,
            final_state_qn = final_state_qn
        )
    ])

    chain2 = MultiChain.from_chains([
        DecayChain(
            topology = topology2,
            resonances = resonances2,
            momenta = momenta,
            final_state_qn = final_state_qn,
        ),
        DecayChain(
            topology = topology2,
            resonances = resonances3,
            momenta = momenta,
            final_state_qn = final_state_qn
        )
    ])

    full = ChainCombiner([chain1, chain2])
    unpolarized, argnames = full.unpolarized_amplitude(full.generate_couplings())
    print(argnames)
    print(unpolarized(*([1] * len(argnames))) )

    polarized, lambdas ,polarized_argnames = full.polarized_amplitude(full.generate_couplings())
    print(lambdas)
    lambda_values = [1, 1, 0, 0]
    print(polarized(*lambda_values,*([1] * len(polarized_argnames))) )

    matrx_function, matrix_argnames = full.matrix_function(full.generate_couplings())
    print(matrix_argnames)
    print(matrx_function(1, *([1] * len(argnames))) )


if __name__ == "__main__":
    threeBodyAmplitude()
    shortThreeBodyAmplitude()
    shortThreeBodyAmplitudeBW()




class AmplitudeParameters:
    def __init__(self, function_arguments:list[str], couplings:dict[tuple[int], str]):
        self.__function_arguments = function_arguments
        self.__couplings = couplings
    
    def decode(arguments=None, couplings=None):
        if arguments is not None and couplings is not None:
            raise ValueError("")


from decayamplitude.particle import Particle, DecaySetup


def replace_in_tupe(tpl, n1, n2):
    """
    Replace all occurrences of n1 with n2 in the nested tuple.
    """
    if not isinstance(tpl, tuple):
        return n2 if tpl == n1 else tpl
    return tuple(replace_in_tupe(x, n1, n2) for x in tpl)

def test_3_body():
    final_state = {
        1: Particle(spin=1, parity=-1, type_id=1),
        2: Particle(spin=0, parity=1),
        3: Particle(spin=1, parity=-1, type_id=1),
    }
    mother = Particle(spin=0, parity=1)

    final_state_no_type_id = {
        1: Particle(spin=1, parity=-1),
        2: Particle(spin=0, parity=1),
        3: Particle(spin=1, parity=-1),
    }

    setup = DecaySetup(final_state_particles=final_state)
    tuples = tuple(
        topo.tuple for topo in setup.topologies
    )
    # ab == ac
    assert len(set(tuples)) - len(set(replace_in_tupe(tuples, 3, 1))) == 1

def test_4_body():
        
    final_state = {
        1: Particle(spin=1, parity=-1, type_id=1),
        2: Particle(spin=0, parity=1),
        3: Particle(spin=1, parity=-1, type_id=1),
        4: Particle(spin=0, parity=1),
    }
    mother = Particle(spin=0, parity=1)

    setup = DecaySetup(final_state_particles=final_state)
    tuples = tuple(
        topo.tuple for topo in setup.topologies
    )
    
    # Removing one final state particle and replacing it with another one
    # Assume 0 -> abcd   and b = c
    # Reduces the amount of unique topologies by 3
    # ((ab), (cd)) == ((ac), (bd)),
    # (((ab) c) d) == (((ac) b) d)
    # ()
    assert len(set(tuples)) - len(set(replace_in_tupe(tuples, 3, 1))) == 5


def test_5_body():
    final_state = {
        1: Particle(spin=1, parity=-1, type_id=1),
        2: Particle(spin=0, parity=1),
        3: Particle(spin=1, parity=-1, type_id=1),
        4: Particle(spin=0, parity=1),
        5: Particle(spin=0, parity=1),
    }
    mother = Particle(spin=0, parity=1)

    setup = DecaySetup(final_state_particles=final_state)
    tuples = tuple(
        topo.tuple for topo in setup.topologies
    )
    
    # Removing one final state particle and replacing it with another one
    # Assume 0 -> abcde   and b = c
    assert len(set(tuples)) - len(set(replace_in_tupe(tuples, 3, 1))) == 42


from typing import Any
from decayamplitude.rotation import QN, Angular
from decayangle.decay_topology import Topology, TopologyCollection


def sorting_function(x, final_state_particles: dict[int, "Particle"]):
    def key(x):
        if not isinstance(x, int):
            return -len(x) * 10000 + key(x[0])
        else:
            return final_state_particles[x].type_id
    if isinstance(x, tuple) or isinstance(x, list):
        return tuple(sorted(x, key=key))
    else:
        return key(x)

class SortingFunction:
    def __init__(self, final_state_particles: dict[int, "Particle"]):
        self.final_state_particles = final_state_particles

    def __call__(self, x):
        return sorting_function(x, self.final_state_particles)

class Particle(QN):
    """
    Particle class allows to store additional attributes to a set of quantum numbers.
    This is meant to be used to describe final state particles. Mainly to handle the case of indistignuishable particles
    in the final state. In this case the ordering of internal nodes is important, since the identical particles must allways be in the same relative order to the other particles in internal nodes.
    Otherwise one needs to apply a transformation to the couplings, wich would be non-trivial to ensure correctness on through the entire decay tree.
    """

    global_names = {}

    @classmethod
    def add_named_partice(cls, name):
        if name in cls.global_names:
            return cls.global_names[name]
        if len(cls.global_names) == 0:
            new_id = 0
        else:
            new_id = max(cls.global_names.values()) + 1
        cls.global_names[name] = new_id
        return new_id

    def __init__(self, spin: Angular | int = None, parity: int = None, quantum_numbers: QN = None, type_id: int = None, name=None):
        if quantum_numbers is None:
            if spin is None or parity is None:
                raise ValueError("Either quantum_numbers or spin and parity must be provided.")
            self.quantum_numbers = QN(angular_momentum=spin, parity=parity)
        else:
            self.quantum_numbers = quantum_numbers
        self.angular = self.quantum_numbers.angular
        self.parity = self.quantum_numbers.parity
        self.type_id = type_id
        self.name = name
        if self.type_id is None and self.name is not None:
            self.type_id = Particle.add_named_partice(self.name)

class DecaySetup:
    def __init__(self, final_state_particles: dict[int, Particle]):
        self.final_state_particles = final_state_particles

        if any(p.type_id for p in self.final_state_particles.values()):
            print(f"At least one particle has a given type_id.")
        for i, p in self.final_state_particles.items():
            if p.type_id is None: 
                p.type_id = i

        self.tc = TopologyCollection(0, list(self.final_state_particles), ordering_function=SortingFunction(self.final_state_particles))
 
    @property
    def topologies(self) -> Topology:
        return self.tc.topologies
    
    def symmetrize(self, topo: Topology) -> list[Topology]:
        return Topology(topo.root.value, topo.tuple, ordering_function=SortingFunction(self.final_state_particles))

    def filled_topologies(self, resonances: dict[tuple[int, ...], Any]):
        toplogies = self.topologies
        def flat(tpl):
            if isinstance(tpl, tuple) or isinstance(tpl, list):
                for item in tpl:
                    yield from flat(item)
            else:
                yield tpl

        def topology_filter(topo: Topology):
            resonances_internal = {
                tuple(sorted(flat(k))): v for k, v in resonances.items()
            }
            topo_nodes = [
                tuple(sorted(flat(node.value))) for node in topo.nodes.values() if not node.final_state
            ]

            return all(
                len(resonances_internal.get(node, [])) > 0 for node in topo_nodes
            )

        return list(
            filter(topology_filter,
                toplogies
            )
        )

